import logging,os,shutil,zipfile
from typing import Dict
from localstack.utils.common import new_tmp_dir,rm_rf
from localstack_ext.bootstrap.pods.constants import COMPRESSION_FORMAT
from localstack_ext.bootstrap.pods.utils.common import PodsConfigContext
LOG=logging.getLogger(__name__)
def extract_meta_and_state_archives(meta_archives,state_archives,config_context):
	G=False;E=config_context;D=meta_archives
	for B in [D,state_archives]:
		for (C,F) in B.items():
			with zipfile.ZipFile(F)as H:
				if B==D:A=E.get_version_meta_archive_path(version=C,with_format=G)
				else:A=E.get_version_state_archive_path(version=C,with_format=G)
				H.extractall(A);shutil.make_archive(base_name=A,format=COMPRESSION_FORMAT,root_dir=A);rm_rf(A);rm_rf(F);LOG.debug(f"Successfully extracted archive {B} for version {C}")
def register_remote(remote_info,config_context):
	B=config_context;A=remote_info
	if B.is_remotely_managed():LOG.warning('Pod is already remotely managed');return
	with open(B.get_remote_info_path(),'w')as C:D=A.get('storage_uuid');E=A.get('qualifying_name');C.write(f"storage_uuid={D}\n");C.write(f"qualifying_name={E}\n")
def merge_version_space(version_space_archive,config_context):
	D=version_space_archive;B=config_context;from localstack_ext.bootstrap.pods.object_storage import get_object_storage_provider as K;L=K(B);E=new_tmp_dir();A=PodsConfigContext(pod_root_dir=E)
	with zipfile.ZipFile(D)as M:M.extractall(A.get_pod_root_dir())
	shutil.copy(A.get_known_ver_path(),B.get_known_ver_path());shutil.copy(A.get_max_ver_path(),B.get_max_ver_path());F=A.get_rev_obj_store_path();N=B.get_rev_obj_store_path()
	for G in os.listdir(F):O=os.path.join(F,G);P=os.path.join(N,G);shutil.copy(O,P)
	for H in os.listdir(A.get_ver_refs_path()):
		I=A.get_version_ref_file_path(H);J=B.get_version_ref_file_path(H)
		with open(I,'r')as Q:C=Q.readline().strip()
		if os.path.isfile(J):L.merge_remote_into_local_version(remote_location=A.get_ver_obj_store_path(),key=C)
		else:R=os.path.join(A.get_ver_obj_store_path(),C);S=os.path.join(B.get_ver_obj_store_path(),C);shutil.copy(I,J);shutil.copy(R,S)
	rm_rf(E);rm_rf(D)


_E='mutables'
_D='service_name'
_C='region'
_B='account_id'
_A='dir_name'
import os,zipfile
from typing import Dict,Protocol
from localstack.utils.common import mkdir,new_tmp_dir
from localstack.utils.generic.singleton_utils import SubtypesInstanceManager
from moto.s3.models import FakeBucket
from moto.sqs.models import Queue
from localstack_ext.bootstrap.pods.models import Serialization
from localstack_ext.bootstrap.pods.service_state import Backends
from localstack_ext.bootstrap.state_merge import merge_kinesis_state,merge_object_state
from localstack_ext.bootstrap.state_merge_ddb import merge_dynamodb
from localstack_ext.bootstrap.state_utils import API_STATES_DIR,DYNAMODB_DIR,KINESIS_DIR,api_states_traverse,load_persisted_object,persist_object
from localstack_ext.utils.persistence import marshal_backend
ROOT_FOLDERS_BY_SERIALIZATION=[API_STATES_DIR,DYNAMODB_DIR,KINESIS_DIR]
class MergeManager(Protocol):
	def merge(A,a,b):...
class DefaultMergeManager(MergeManager):
	def merge(F,a,b):
		from localstack_ext.utils.persistence import unmarshal_backend as B
		for A in b:
			if A in a:C=B(b[A]);D=B(a[A]);E=merge_object_state(D,C);a[A]=marshal_backend(E)
			else:a[A]=b[A]
class DynamoDBMergeManager(MergeManager):
	def merge(A,a,b):...
class KinesisMergeManager(MergeManager):
	def merge(A,a,b):...
def get_merge_manager(service):
	A=service
	if A=='dynamodb':return DynamoDBMergeManager()
	if A=='kinesis':return KinesisMergeManager()
	return DefaultMergeManager()
class CloudPodsMergeManager(SubtypesInstanceManager):
	@staticmethod
	def _is_special_case(obj):return isinstance(obj,(Queue,FakeBucket))
	def two_way_merge(A,from_state_files,to_state_files):raise NotImplementedError
	def three_way_merge(A,common_ancestor_state_files,from_state_files,to_state_files):raise NotImplementedError
class CloudPodsMergeManagerDynamoDB(CloudPodsMergeManager):
	@staticmethod
	def impl_name():return Serialization.DDB.value
	def two_way_merge(A,from_state_files,to_state_files):merge_dynamodb(from_state_files,to_state_files)
	def three_way_merge(A,common_ancestor_state_files,from_state_files,to_state_files):merge_dynamodb(from_state_files,to_state_files)
class CloudPodsMergeManagerKinesis(CloudPodsMergeManager):
	@staticmethod
	def impl_name():return Serialization.KINESIS.value
	def two_way_merge(A,from_state_files,to_state_files):merge_kinesis_state(to_state_files,from_state_files)
	def three_way_merge(A,common_ancestor_state_files,from_state_files,to_state_files):A.two_way_merge(to_state_files,from_state_files)
class CloudPodsMergeManagerMain(CloudPodsMergeManager):
	@staticmethod
	def impl_name():return Serialization.MAIN.value
	@staticmethod
	def _merge_three_way_dir_func(**A):
		K=A.get(_A);E=A.get(_B);C=A.get('fname');F=A.get(_C);G=A.get(_D);H=A.get(_E);L=H[0];M=H[1];N=os.path.join(K,C);O=os.path.join(M,E,G,F);P=os.path.join(O,C);I=os.path.join(L,E,G,F);B=os.path.join(I,C);D=load_persisted_object(N);Q=load_persisted_object(P);R=CloudPodsMergeManager._is_special_case(D)
		if os.path.isfile(B):
			if not R:J=load_persisted_object(B);merge_object_state(J,D,Q);persist_object(J,B)
		else:mkdir(I);persist_object(D,B)
	@staticmethod
	def _merge_two_state_dir_func(**A):
		G=A.get(_A);H=A.get(_B);D=A.get('fname');I=A.get(_C);J=A.get(_D);K=A.get(_E);L=K[0];M=os.path.join(G,D);E=os.path.join(L,H,J,I);B=os.path.join(E,D);C=load_persisted_object(M);N=CloudPodsMergeManager._is_special_case(C)
		if os.path.isfile(B):
			if not N:F=load_persisted_object(B);merge_object_state(F,C);persist_object(F,B)
		else:mkdir(E);persist_object(C,B)
	def two_way_merge(A,from_state_files,to_state_files):api_states_traverse(api_states_path=to_state_files,side_effect=CloudPodsMergeManagerMain._merge_two_state_dir_func,mutables=[from_state_files])
	def three_way_merge(A,common_ancestor_state_files,from_state_files,to_state_files):api_states_traverse(api_states_path=to_state_files,side_effect=CloudPodsMergeManagerMain._merge_three_way_dir_func,mutables=[from_state_files,common_ancestor_state_files])
def create_tmp_archives_by_serialization_mechanism(archive_dir):
	A=new_tmp_dir()
	with zipfile.ZipFile(archive_dir)as E:E.extractall(A)
	B={'root':A}
	for C in ROOT_FOLDERS_BY_SERIALIZATION:D=os.path.join(A,C);mkdir(D);B[C]=D
	return B


_A='BackupPlanId'
from localstack.services.cloudformation.service_models import GenericBaseModel
from localstack.utils.aws import aws_stack
class BackupPlan(GenericBaseModel):
	@staticmethod
	def cloudformation_type():return'AWS::Backup::BackupPlan'
	def fetch_state(A,stack_name,resources):B='BackupPlanName';C=aws_stack.connect_to_service('backup');D=C.list_backup_plans().get('BackupPlansList',[]);E=A.resolve_refs_recursively(stack_name,A.props[B],resources);F=[A for A in D if A[B]==E];return(F or[None])[0]
	def get_physical_resource_id(A,attribute,**B):return A.props.get(_A)
	@staticmethod
	def get_deploy_templates():A='function';return{'create':{A:'create_backup_plan'},'delete':{A:'delete_backup_plan','parameters':[_A]}}


from localstack.aws.protocol import service_router as localstack_service_router
from localstack.utils.patch import patch
def patch_service_router():
	C='appsync';import re;from typing import Callable,Optional;from localstack.http import Request;from localstack.services.apigateway.context import ApiInvocationContext as F;from localstack_ext.services.apigateway.apigateway_extended import is_custom_domain_api_invocation as G
	@patch(localstack_service_router.custom_signing_name_rules)
	def A(fn,signing_name,path,**C):
		B='rds';A=signing_name
		if A in[B,'docdb','neptune']:return B
		return fn(A,path,**C)
	@patch(localstack_service_router.custom_host_addressing_rules)
	def B(fn,host,**B):
		A=host
		if'.cloudfront.'in A:return'cloudfront'
		if'mediastore-'in A:return'mediastore-data'
		if'.appsync-api.'in A:return C
		return fn(A,**B)
	@patch(localstack_service_router.legacy_rules)
	def D(fn,request,**H):
		E='apigateway';A=request;B=A.path;I=A.method;J=A.data;D=A.headers;K=F(I,B,data=J,headers=D)
		if G(K):return E
		if re.match('/graphql/[a-zA-Z0-9-]+',B):return C
		L=D.get('authorization','')
		if L.startswith('Bearer '):return E
		if B.startswith('/_messages_'):return'ses'
		if'/2018-06-01/runtime'in B:return'lambda'
		return fn(request=A,**H)


_A=False
from datetime import datetime
from enum import Enum
from typing import Set
from localstack_ext.bootstrap.pods.constants import COMMIT_TXT_LAYOUT,REV_TXT_LAYOUT,STATE_TXT_LAYOUT,STATE_TXT_METADATA,VER_TXT_LAYOUT
from localstack_ext.bootstrap.state_utils import API_STATES_DIR,DYNAMODB_DIR,KINESIS_DIR
class PodObject:
	def __init__(A,hash_ref):A.hash_ref=hash_ref
class Serialization(Enum):MAIN=API_STATES_DIR;DDB=DYNAMODB_DIR;KINESIS=KINESIS_DIR;serializer_root_lookup={str(MAIN):API_STATES_DIR,str(DDB):DYNAMODB_DIR,str(KINESIS):KINESIS_DIR}
class StateFileRef(PodObject):
	txt_layout=STATE_TXT_LAYOUT;metadata_layout=STATE_TXT_METADATA
	def __init__(A,hash_ref,rel_path,file_name,size,service,region,account_id,serialization):super(StateFileRef,A).__init__(hash_ref);A.rel_path=rel_path;A.file_name=file_name;A.size=size;A.service=service;A.region=region;A.account_id=account_id;A.serialization=serialization
	def __str__(A):return A.txt_layout.format(size=A.size,account_id=A.account_id,service=A.service,region=A.region,hash_ref=A.hash_ref,file_name=A.file_name,rel_path=A.rel_path,serialization=A.serialization)
	def __eq__(A,other):
		B=other
		if not B:return _A
		if not isinstance(B,StateFileRef):return _A
		return A.hash_ref==B.hash_ref and A.account_id==B.account_id and A.region==B.region and A.service==A.service and A.file_name==B.file_name and A.size==B.size
	def __hash__(A):return hash((A.hash_ref,A.account_id,A.region,A.service,A.file_name,A.size))
	def congruent(A,other):
		B=other
		if not B:return _A
		if not isinstance(B,StateFileRef):return _A
		return A.region==B.region and A.account_id==A.account_id and A.service==B.service and A.file_name==B.file_name and A.rel_path==B.rel_path
	def any_congruence(A,others):
		for B in others:
			if A.congruent(B):return True
		return _A
	def metadata(A):return A.metadata_layout.format(size=A.size,service=A.service,region=A.region)
class PodNode(PodObject):
	def __init__(A,hash_ref,state_files,parent_ptr):super(PodNode,A).__init__(hash_ref);A.state_files=state_files;A.parent_ptr=parent_ptr
	def state_files_info(A):return '\n'.join(list(map(lambda state_file:str(state_file),A.state_files)))
class Commit:
	txt_layout=COMMIT_TXT_LAYOUT
	def __init__(A,tail_ptr,head_ptr,message,timestamp=str(datetime.now().timestamp()),delta_log_ptr=None):A.tail_ptr=tail_ptr;A.head_ptr=head_ptr;A.message=message;A.timestamp=timestamp;A.delta_log_ptr=delta_log_ptr
	def __str__(A):return A.txt_layout.format(tail_ptr=A.tail_ptr,head_ptr=A.head_ptr,message=A.message,timestamp=A.timestamp,log_hash=A.delta_log_ptr)
	def info_str(A,from_node,to_node):return f"from: {from_node}, to: {to_node}, message: {A.message}, time: {datetime.fromtimestamp(float(A.timestamp))}"
class Revision(PodNode):
	txt_layout=REV_TXT_LAYOUT
	def __init__(A,hash_ref,state_files,parent_ptr,creator,rid,revision_number,assoc_commit=None):super(Revision,A).__init__(hash_ref,state_files,parent_ptr);A.creator=creator;A.rid=rid;A.revision_number=revision_number;A.assoc_commit=assoc_commit
	def __str__(A):return A.txt_layout.format(hash_ref=A.hash_ref,parent=A.parent_ptr,creator=A.creator,rid=A.rid,rev_no=A.revision_number,state_files=';'.join(map(lambda state_file:str(state_file),A.state_files))if A.state_files else'',assoc_commit=A.assoc_commit)
class Version(PodNode):
	txt_layout=VER_TXT_LAYOUT
	def __init__(A,hash_ref,state_files,parent_ptr,creator,comment,active_revision_ptr,outgoing_revision_ptrs,incoming_revision_ptr,version_number):super(Version,A).__init__(hash_ref,state_files,parent_ptr);A.creator=creator;A.comment=comment;A.active_revision_ptr=active_revision_ptr;A.outgoing_revision_ptrs=outgoing_revision_ptrs;A.incoming_revision_ptr=incoming_revision_ptr;A.version_number=version_number
	def __str__(A):return VER_TXT_LAYOUT.format(hash_ref=A.hash_ref,parent=A.parent_ptr,creator=A.creator,comment=A.comment,version_number=A.version_number,active_revision=A.active_revision_ptr,outgoing_revisions=';'.join(A.outgoing_revision_ptrs),incoming_revision=A.incoming_revision_ptr,state_files=';'.join(map(lambda stat_file:str(stat_file),A.state_files))if A.state_files else'')
	def info_str(A):return f"{A.version_number}, {A.creator}, {A.comment}"


import sys
from typing import List, Optional

if sys.version_info >= (3, 8):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict

from localstack.aws.api import RequestContext, ServiceException, ServiceRequest, handler

MaxResults = int
NextToken = str
PageSize = int
Qos = int
Retain = bool
ShadowName = str
ThingName = str
Topic = str
errorMessage = str


class ConflictException(ServiceException):
    """The specified version does not match the version of the document."""

    message: Optional[errorMessage]


class InternalFailureException(ServiceException):
    """An unexpected error has occurred."""

    message: Optional[errorMessage]


class InvalidRequestException(ServiceException):
    """The request is not valid."""

    message: Optional[errorMessage]


class MethodNotAllowedException(ServiceException):
    """The specified combination of HTTP verb and URI is not supported."""

    message: Optional[errorMessage]


class RequestEntityTooLargeException(ServiceException):
    """The payload exceeds the maximum size allowed."""

    message: Optional[errorMessage]


class ResourceNotFoundException(ServiceException):
    """The specified resource does not exist."""

    message: Optional[errorMessage]


class ServiceUnavailableException(ServiceException):
    """The service is temporarily unavailable."""

    message: Optional[errorMessage]


class ThrottlingException(ServiceException):
    """The rate exceeds the limit."""

    message: Optional[errorMessage]


class UnauthorizedException(ServiceException):
    """You are not authorized to perform this operation."""

    message: Optional[errorMessage]


class UnsupportedDocumentEncodingException(ServiceException):
    """The document encoding is not supported."""

    message: Optional[errorMessage]


class DeleteThingShadowRequest(ServiceRequest):
    """The input for the DeleteThingShadow operation."""

    thingName: ThingName
    shadowName: Optional[ShadowName]


JsonDocument = bytes


class DeleteThingShadowResponse(TypedDict, total=False):
    """The output from the DeleteThingShadow operation."""

    payload: JsonDocument


class GetRetainedMessageRequest(ServiceRequest):
    """The input for the GetRetainedMessage operation."""

    topic: Topic


Timestamp = int
Payload = bytes


class GetRetainedMessageResponse(TypedDict, total=False):
    """The output from the GetRetainedMessage operation."""

    topic: Optional[Topic]
    payload: Optional[Payload]
    qos: Optional[Qos]
    lastModifiedTime: Optional[Timestamp]


class GetThingShadowRequest(ServiceRequest):
    """The input for the GetThingShadow operation."""

    thingName: ThingName
    shadowName: Optional[ShadowName]


class GetThingShadowResponse(TypedDict, total=False):
    """The output from the GetThingShadow operation."""

    payload: Optional[JsonDocument]


class ListNamedShadowsForThingRequest(ServiceRequest):
    thingName: ThingName
    nextToken: Optional[NextToken]
    pageSize: Optional[PageSize]


NamedShadowList = List[ShadowName]


class ListNamedShadowsForThingResponse(TypedDict, total=False):
    results: Optional[NamedShadowList]
    nextToken: Optional[NextToken]
    timestamp: Optional[Timestamp]


class ListRetainedMessagesRequest(ServiceRequest):
    nextToken: Optional[NextToken]
    maxResults: Optional[MaxResults]


PayloadSize = int


class RetainedMessageSummary(TypedDict, total=False):
    """Information about a single retained message."""

    topic: Optional[Topic]
    payloadSize: Optional[PayloadSize]
    qos: Optional[Qos]
    lastModifiedTime: Optional[Timestamp]


RetainedMessageList = List[RetainedMessageSummary]


class ListRetainedMessagesResponse(TypedDict, total=False):
    retainedTopics: Optional[RetainedMessageList]
    nextToken: Optional[NextToken]


class PublishRequest(ServiceRequest):
    """The input for the Publish operation."""

    topic: Topic
    qos: Optional[Qos]
    retain: Optional[Retain]
    payload: Optional[Payload]


class UpdateThingShadowRequest(ServiceRequest):
    """The input for the UpdateThingShadow operation."""

    thingName: ThingName
    shadowName: Optional[ShadowName]
    payload: JsonDocument


class UpdateThingShadowResponse(TypedDict, total=False):
    """The output from the UpdateThingShadow operation."""

    payload: Optional[JsonDocument]


class IotDataApi:

    service = "iot-data"
    version = "2015-05-28"

    @handler("DeleteThingShadow")
    def delete_thing_shadow(
        self, context: RequestContext, thing_name: ThingName, shadow_name: ShadowName = None
    ) -> DeleteThingShadowResponse:
        """Deletes the shadow for the specified thing.

        Requires permission to access the
        `DeleteThingShadow <https://docs.aws.amazon.com/service-authorization/latest/reference/list_awsiot.html#awsiot-actions-as-permissions>`__
        action.

        For more information, see
        `DeleteThingShadow <http://docs.aws.amazon.com/iot/latest/developerguide/API_DeleteThingShadow.html>`__
        in the IoT Developer Guide.

        :param thing_name: The name of the thing.
        :param shadow_name: The name of the shadow.
        :returns: DeleteThingShadowResponse
        :raises ResourceNotFoundException:
        :raises InvalidRequestException:
        :raises ThrottlingException:
        :raises UnauthorizedException:
        :raises ServiceUnavailableException:
        :raises InternalFailureException:
        :raises MethodNotAllowedException:
        :raises UnsupportedDocumentEncodingException:
        """
        raise NotImplementedError

    @handler("GetRetainedMessage")
    def get_retained_message(
        self, context: RequestContext, topic: Topic
    ) -> GetRetainedMessageResponse:
        """Gets the details of a single retained message for the specified topic.

        This action returns the message payload of the retained message, which
        can incur messaging costs. To list only the topic names of the retained
        messages, call
        `ListRetainedMessages </iot/latest/developerguide/API_iotdata_ListRetainedMessages.html>`__.

        Requires permission to access the
        `GetRetainedMessage <https://docs.aws.amazon.com/service-authorization/latest/reference/list_awsiotfleethubfordevicemanagement.html#awsiotfleethubfordevicemanagement-actions-as-permissions>`__
        action.

        For more information about messaging costs, see `Amazon Web Services IoT
        Core pricing -
        Messaging <http://aws.amazon.com/iot-core/pricing/#Messaging>`__.

        :param topic: The topic name of the retained message to retrieve.
        :returns: GetRetainedMessageResponse
        :raises InvalidRequestException:
        :raises ResourceNotFoundException:
        :raises ThrottlingException:
        :raises UnauthorizedException:
        :raises ServiceUnavailableException:
        :raises InternalFailureException:
        :raises MethodNotAllowedException:
        """
        raise NotImplementedError

    @handler("GetThingShadow")
    def get_thing_shadow(
        self, context: RequestContext, thing_name: ThingName, shadow_name: ShadowName = None
    ) -> GetThingShadowResponse:
        """Gets the shadow for the specified thing.

        Requires permission to access the
        `GetThingShadow <https://docs.aws.amazon.com/service-authorization/latest/reference/list_awsiot.html#awsiot-actions-as-permissions>`__
        action.

        For more information, see
        `GetThingShadow <http://docs.aws.amazon.com/iot/latest/developerguide/API_GetThingShadow.html>`__
        in the IoT Developer Guide.

        :param thing_name: The name of the thing.
        :param shadow_name: The name of the shadow.
        :returns: GetThingShadowResponse
        :raises InvalidRequestException:
        :raises ResourceNotFoundException:
        :raises ThrottlingException:
        :raises UnauthorizedException:
        :raises ServiceUnavailableException:
        :raises InternalFailureException:
        :raises MethodNotAllowedException:
        :raises UnsupportedDocumentEncodingException:
        """
        raise NotImplementedError

    @handler("ListNamedShadowsForThing")
    def list_named_shadows_for_thing(
        self,
        context: RequestContext,
        thing_name: ThingName,
        next_token: NextToken = None,
        page_size: PageSize = None,
    ) -> ListNamedShadowsForThingResponse:
        """Lists the shadows for the specified thing.

        Requires permission to access the
        `ListNamedShadowsForThing <https://docs.aws.amazon.com/service-authorization/latest/reference/list_awsiot.html#awsiot-actions-as-permissions>`__
        action.

        :param thing_name: The name of the thing.
        :param next_token: The token to retrieve the next set of results.
        :param page_size: The result page size.
        :returns: ListNamedShadowsForThingResponse
        :raises ResourceNotFoundException:
        :raises InvalidRequestException:
        :raises ThrottlingException:
        :raises UnauthorizedException:
        :raises ServiceUnavailableException:
        :raises InternalFailureException:
        :raises MethodNotAllowedException:
        """
        raise NotImplementedError

    @handler("ListRetainedMessages")
    def list_retained_messages(
        self, context: RequestContext, next_token: NextToken = None, max_results: MaxResults = None
    ) -> ListRetainedMessagesResponse:
        """Lists summary information about the retained messages stored for the
        account.

        This action returns only the topic names of the retained messages. It
        doesn't return any message payloads. Although this action doesn't return
        a message payload, it can still incur messaging costs.

        To get the message payload of a retained message, call
        `GetRetainedMessage <https://docs.aws.amazon.com/iot/latest/developerguide/API_iotdata_GetRetainedMessage.html>`__
        with the topic name of the retained message.

        Requires permission to access the
        `ListRetainedMessages <https://docs.aws.amazon.com/service-authorization/latest/reference/list_awsiotfleethubfordevicemanagement.html#awsiotfleethubfordevicemanagement-actions-as-permissions>`__
        action.

        For more information about messaging costs, see `Amazon Web Services IoT
        Core pricing -
        Messaging <http://aws.amazon.com/iot-core/pricing/#Messaging>`__.

        :param next_token: To retrieve the next set of results, the ``nextToken`` value from a
        previous response; otherwise **null** to receive the first set of
        results.
        :param max_results: The maximum number of results to return at one time.
        :returns: ListRetainedMessagesResponse
        :raises InvalidRequestException:
        :raises ThrottlingException:
        :raises UnauthorizedException:
        :raises ServiceUnavailableException:
        :raises InternalFailureException:
        :raises MethodNotAllowedException:
        """
        raise NotImplementedError

    @handler("Publish")
    def publish(
        self,
        context: RequestContext,
        topic: Topic,
        qos: Qos = None,
        retain: Retain = None,
        payload: Payload = None,
    ) -> None:
        """Publishes an MQTT message.

        Requires permission to access the
        `Publish <https://docs.aws.amazon.com/service-authorization/latest/reference/list_awsiot.html#awsiot-actions-as-permissions>`__
        action.

        For more information about MQTT messages, see `MQTT
        Protocol <http://docs.aws.amazon.com/iot/latest/developerguide/mqtt.html>`__
        in the IoT Developer Guide.

        For more information about messaging costs, see `Amazon Web Services IoT
        Core pricing -
        Messaging <http://aws.amazon.com/iot-core/pricing/#Messaging>`__.

        :param topic: The name of the MQTT topic.
        :param qos: The Quality of Service (QoS) level.
        :param retain: A Boolean value that determines whether to set the RETAIN flag when the
        message is published.
        :param payload: The message body.
        :raises InternalFailureException:
        :raises InvalidRequestException:
        :raises UnauthorizedException:
        :raises MethodNotAllowedException:
        """
        raise NotImplementedError

    @handler("UpdateThingShadow")
    def update_thing_shadow(
        self,
        context: RequestContext,
        thing_name: ThingName,
        payload: JsonDocument,
        shadow_name: ShadowName = None,
    ) -> UpdateThingShadowResponse:
        """Updates the shadow for the specified thing.

        Requires permission to access the
        `UpdateThingShadow <https://docs.aws.amazon.com/service-authorization/latest/reference/list_awsiot.html#awsiot-actions-as-permissions>`__
        action.

        For more information, see
        `UpdateThingShadow <http://docs.aws.amazon.com/iot/latest/developerguide/API_UpdateThingShadow.html>`__
        in the IoT Developer Guide.

        :param thing_name: The name of the thing.
        :param payload: The state information, in JSON format.
        :param shadow_name: The name of the shadow.
        :returns: UpdateThingShadowResponse
        :raises ConflictException:
        :raises RequestEntityTooLargeException:
        :raises InvalidRequestException:
        :raises ThrottlingException:
        :raises UnauthorizedException:
        :raises ServiceUnavailableException:
        :raises InternalFailureException:
        :raises MethodNotAllowedException:
        :raises UnsupportedDocumentEncodingException:
        """
        raise NotImplementedError



_S='RouteResponseId'
_R='VpcLinkId'
_Q='DeploymentId'
_P='IntegrationId'
_O='RouteKey'
_N='Body'
_M='RouteId'
_L='Description'
_K='DomainName'
_J='StageName'
_I='delete'
_H='Name'
_G='parameters'
_F='Items'
_E='create'
_D='apigatewayv2'
_C='ApiId'
_B=None
_A='function'
import json
from localstack.services.cloudformation.service_models import REF_ID_ATTRS,GenericBaseModel
from localstack.utils.aws import aws_stack
from localstack.utils.common import select_attributes,short_uid
class ApiGatewayV2VpcLink(GenericBaseModel):
	@staticmethod
	def cloudformation_type():return'AWS::ApiGatewayV2::VpcLink'
	def fetch_state(A,stack_name,resources):B=aws_stack.connect_to_service(_D);C=B.get_vpc_links()[_F];D=A.resolve_refs_recursively(stack_name,A.props[_H],resources);E=[A for A in C if A[_H]==D];return(E or[_B])[0]
	def get_physical_resource_id(A,attribute,**B):return A.props.get(_R)
	@staticmethod
	def get_deploy_templates():return{_E:{_A:'create_vpc_link'},_I:{_A:'delete_vpc_link',_G:[_R]}}
class ApiGatewayV2DomainName(GenericBaseModel):
	@staticmethod
	def cloudformation_type():return'AWS::ApiGatewayV2::DomainName'
	def fetch_state(A,stack_name,resources):B=aws_stack.connect_to_service(_D);C=B.get_domain_names()[_F];D=A.resolve_refs_recursively(stack_name,A.props[_K],resources);E=([A for A in C if A[_K]==D]or[_B])[0];return E
	def get_physical_resource_id(A,attribute,**B):return A.props.get(_K)
	@staticmethod
	def get_deploy_templates():return{_E:{_A:'create_domain_name'},_I:{_A:'delete_domain_name',_G:[_K]}}
class ApiGatewayV2Authorizer(GenericBaseModel):
	@staticmethod
	def cloudformation_type():return'AWS::ApiGatewayV2::Authorizer'
	def fetch_state(A,stack_name,resources):C=resources;B=stack_name;E=aws_stack.connect_to_service(_D);D=A.props;F=A.resolve_refs_recursively(B,D[_C],C);G=A.resolve_refs_recursively(B,D[_H],C);H=E.get_authorizers(ApiId=F)[_F];I=([A for A in H if A[_H]==G]or[_B])[0];return I
	def get_physical_resource_id(A,attribute,**B):return A.props.get('AuthorizerId')
	@staticmethod
	def get_deploy_templates():return{_E:{_A:'create_authorizer'}}
class ApiGatewayV2Api(GenericBaseModel):
	@staticmethod
	def cloudformation_type():return'AWS::ApiGatewayV2::Api'
	def fetch_state(B,stack_name,resources):A=B.props;C=A[_N].get('info',{}).get('title')if A.get(_N)else A[_H];D=aws_stack.connect_to_service(_D);E=D.get_apis()[_F];F=([A for A in E if A[_H]==C]or[_B])[0];return F
	def get_physical_resource_id(A,attribute=_B,**B):
		if attribute in REF_ID_ATTRS:return A.props.get(_C)
	@staticmethod
	def get_deploy_templates():
		def A(resource_id,resources,*I):
			E='Basepath';F=resources[resource_id];B=F['Properties'];C=aws_stack.connect_to_service(_D);A=B.get(_N)
			if A:D=B.get(E);A.setdefault('info',{}).setdefault('title','api-%s'%short_uid());A=json.dumps(A);G={E:D}if D else{};return C.import_api(Body=A,**G)
			H=select_attributes(B,['ApiKeySelectionExpression','CorsConfiguration','CredentialsArn',_L,'DisableSchemaValidation',_H,'ProtocolType',_O,'RouteSelectionExpression','Tags','Target','Version']);return C.create_api(**H)
		return{_E:{_A:A}}
class ApiGatewayV2IntegrationResponse(GenericBaseModel):
	@staticmethod
	def cloudformation_type():return'AWS::ApiGatewayV2::IntegrationResponse'
	def fetch_state(A,stack_name,resources):E='IntegrationResponseKey';C=resources;B=stack_name;F=aws_stack.connect_to_service(_D);D=A.props;G=A.resolve_refs_recursively(B,D[_C],C);H=A.resolve_refs_recursively(B,D[_P],C);I=A.resolve_refs_recursively(B,D[E],C);J=F.get_integration_responses(ApiId=G,IntegrationId=H).get(_F,[]);K=[A for A in J if A[E]==I];return(K or[_B])[0]
	def get_physical_resource_id(A,attribute=_B,**B):return A.props.get('IntegrationResponseId')
	def get_cfn_attribute(B,attribute_name):
		A=attribute_name
		if A in REF_ID_ATTRS:return B.get_physical_resource_id(A)
		return super(ApiGatewayV2IntegrationResponse,B).get_cfn_attribute(A)
	@staticmethod
	def get_deploy_templates():return{_E:{_A:'create_integration_response'}}
class ApiGatewayV2Integration(GenericBaseModel):
	@staticmethod
	def cloudformation_type():return'AWS::ApiGatewayV2::Integration'
	def get_physical_resource_id(A,attribute,**B):
		if attribute in REF_ID_ATTRS:return A.props.get(_P)
	def fetch_state(A,stack_name,resources):G='IntegrationMethod';F='IntegrationUri';E='IntegrationType';C=resources;B=stack_name;H=aws_stack.connect_to_service(_D);D=A.props;I=A.resolve_refs_recursively(B,D.get(_C),C);J=A.resolve_refs_recursively(B,D.get(E),C);K=A.resolve_refs_recursively(B,D.get(F),C);L=A.resolve_refs_recursively(B,D.get(G),C);M=H.get_integrations(ApiId=I);N=[A for A in M.get(_F,[])if J==A.get(E)and K in[_B,A.get(F)]and L in[_B,A.get(G)]];return(N or[_B])[0]
	@staticmethod
	def get_deploy_templates():return{_E:{_A:'create_integration'},_I:{_A:'delete_integration',_G:[_C,_P]}}
class ApiGatewayV2Deployment(GenericBaseModel):
	@staticmethod
	def cloudformation_type():return'AWS::ApiGatewayV2::Deployment'
	def get_physical_resource_id(A,attribute,**B):
		if attribute in REF_ID_ATTRS:return A.props.get(_Q)
	def fetch_state(A,stack_name,resources):C=resources;B=stack_name;D=A.props;F=A.resolve_refs_recursively(B,D.get(_C),C);G=A.resolve_refs_recursively(B,D.get(_J),C);H=A.resolve_refs_recursively(B,D.get(_L),C);I=aws_stack.connect_to_service(_D);E=I.get_deployments(ApiId=F)[_F];E=[A for A in E if A.get(_J)==G or A.get(_L)==H];return(E or[_B])[0]
	@staticmethod
	def get_deploy_templates():return{_E:{_A:'create_deployment'},_I:{_A:'delete_deployment',_G:[_C,_Q]}}
class ApiGatewayV2Stage(GenericBaseModel):
	@staticmethod
	def cloudformation_type():return'AWS::ApiGatewayV2::Stage'
	def fetch_state(A,stack_name,resources):C=resources;B=stack_name;D=A.props;E=A.resolve_refs_recursively(B,D.get(_C),C);F=A.resolve_refs_recursively(B,D.get(_J),C);G=aws_stack.connect_to_service(_D);return G.get_stage(ApiId=E,StageName=F)or _B
	def get_physical_resource_id(A,attribute=_B,**B):
		if attribute in REF_ID_ATTRS:return A.props.get(_J)
	@staticmethod
	def get_deploy_templates():return{_E:{_A:'create_stage',_G:[_C,_Q,_J,_L]},_I:{_A:'delete_stage',_G:[_C,_J]}}
class ApiGatewayV2RouteResponse(GenericBaseModel):
	@staticmethod
	def cloudformation_type():return'AWS::ApiGatewayV2::RouteResponse'
	def fetch_state(A,stack_name,resources):E='RouteResponseKey';C=resources;B=stack_name;F=aws_stack.connect_to_service(_D);D=A.props;G=A.resolve_refs_recursively(B,D[_C],C);H=A.resolve_refs_recursively(B,D[_M],C);I=A.resolve_refs_recursively(B,D[E],C);J=F.get_route_responses(ApiId=G,RouteId=H).get(_F,[]);K=[A for A in J if A[E]==I];return(K or[_B])[0]
	def get_physical_resource_id(A,attribute,**B):return A.props.get(_S)
	@staticmethod
	def get_deploy_templates():return{_E:{_A:'create_route_response'},_I:{_A:'delete_route_response',_G:[_C,_M,_S]}}
class ApiGatewayV2Route(GenericBaseModel):
	@staticmethod
	def cloudformation_type():return'AWS::ApiGatewayV2::Route'
	def fetch_state(A,stack_name,resources):C=resources;B=stack_name;E=aws_stack.connect_to_service(_D);D=A.props;F=A.resolve_refs_recursively(B,D.get(_C),C);G=A.resolve_refs_recursively(B,D.get(_O),C);H=E.get_routes(ApiId=F).get(_F,[]);I=[A for A in H if A[_O]==G];return(I or[_B])[0]
	def get_physical_resource_id(A,attribute=_B,**B):return A.props.get(_M)
	@staticmethod
	def get_deploy_templates():return{_E:{_A:'create_route'},_I:{_A:'delete_route',_G:[_C,_M]}}


import abc,logging,os
from typing import Dict,Optional,Set,Union
from localstack_ext.bootstrap.pods.models import Commit,PodNode,PodObject,Revision,StateFileRef,Version
LOG=logging.getLogger(__name__)
class PodsSerializer(abc.ABC):
	def __init__(A):0
	@abc.abstractmethod
	def store_obj(self,pod_object,path):0
	@abc.abstractmethod
	def retrieve_obj(self,key,remote_path,local_path):0
	@staticmethod
	def _deserialize_state_files(state_files_str):
		B=state_files_str
		if not B:return set()
		D=B.split(';');C=set()
		for E in D:A=list(map(lambda x:x.split(':')[1],E.split(',')));C.add(StateFileRef(size=int(A[0]),service=A[1],account_id=A[2],region=A[3],hash_ref=A[4],file_name=A[5],rel_path=A[6],serialization=A[7]))
		return C
class VersionSerializerTxt(PodsSerializer):
	def store_obj(C,pod_object,path):
		A=pod_object
		with open(os.path.join(path,A.hash_ref),'w')as B:B.write(str(A))
		return A.hash_ref
	def retrieve_obj(D,key,remote_path,local_path):
		C=remote_path
		if C:B=os.path.join(C,key)
		else:B=os.path.join(local_path,key)
		if not os.path.isfile(B):LOG.debug(f"No Version Obj file found in path {B}");return
		with open(B,'r')as E:F=list(map(lambda line:line.rstrip(),E.readlines()));A=list(map(lambda line:line.split('=')[1],F));G=D._deserialize_state_files(A[8]);return Version(parent_ptr=A[0],hash_ref=A[1],creator=A[2],comment=A[3],version_number=int(A[4]),active_revision_ptr=A[5],outgoing_revision_ptrs=set(A[6].split(';')),incoming_revision_ptr=A[7],state_files=G)
class RevisionSerializerTxt(PodsSerializer):
	def store_obj(D,pod_obj,path):
		A=pod_obj
		try:
			with open(os.path.join(path,A.hash_ref),'w')as B:B.write(str(A))
		except FileNotFoundError as C:print(C)
		return A.hash_ref
	def retrieve_obj(C,key,remote_path,local_path):
		B=os.path.join(local_path,key)
		if not os.path.isfile(B):LOG.debug(f"No Revision Obj file found in path {B}");return
		def D(commit_str):
			B=commit_str
			if not B or B=='None':return
			A=list(map(lambda commit_attr:commit_attr.split(':')[1],B.split(',')));return Commit(tail_ptr=A[0],head_ptr=A[1],message=A[2],timestamp=A[3],delta_log_ptr=A[4])
		with open(B)as E:F=list(map(lambda line:line.rstrip(),E.readlines()));A=list(map(lambda line:line.split('=')[1],F));G=C._deserialize_state_files(A[5]);return Revision(parent_ptr=A[0],hash_ref=A[1],creator=A[2],rid=A[3],revision_number=int(A[4]),state_files=G,assoc_commit=D(A[6]))
version_serializer=VersionSerializerTxt()
revision_serializer=RevisionSerializerTxt()
txt_serializers={'version':version_serializer,'revision':revision_serializer}


import sys
from datetime import datetime
from typing import Optional

if sys.version_info >= (3, 8):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict

from localstack.aws.api import RequestContext, ServiceException, ServiceRequest, handler

_string = str


class ForbiddenException(ServiceException):
    """The caller is not authorized to invoke this operation."""


class GoneException(ServiceException):
    """The connection with the provided id no longer exists."""


class PayloadTooLargeException(ServiceException):
    """The data has exceeded the maximum size allowed."""

    Message: Optional[_string]


class LimitExceededException(ServiceException):
    """The client is sending more than the allowed number of requests per unit
    of time or the WebSocket client side buffer is full.
    """


Data = bytes


class DeleteConnectionRequest(ServiceRequest):
    ConnectionId: _string


class GetConnectionRequest(ServiceRequest):
    ConnectionId: _string


_timestampIso8601 = datetime


class Identity(TypedDict, total=False):
    SourceIp: _string
    UserAgent: _string


class GetConnectionResponse(TypedDict, total=False):
    ConnectedAt: Optional[_timestampIso8601]
    Identity: Optional[Identity]
    LastActiveAt: Optional[_timestampIso8601]


class PostToConnectionRequest(ServiceRequest):
    Data: Data
    ConnectionId: _string


class ApigatewaymanagementapiApi:

    service = "apigatewaymanagementapi"
    version = "2018-11-29"

    @handler("DeleteConnection")
    def delete_connection(self, context: RequestContext, connection_id: _string) -> None:
        """Delete the connection with the provided id.

        :param connection_id: .
        :raises GoneException:
        :raises LimitExceededException:
        :raises ForbiddenException:
        """
        raise NotImplementedError

    @handler("GetConnection")
    def get_connection(
        self, context: RequestContext, connection_id: _string
    ) -> GetConnectionResponse:
        """Get information about the connection with the provided id.

        :param connection_id: .
        :returns: GetConnectionResponse
        :raises GoneException:
        :raises LimitExceededException:
        :raises ForbiddenException:
        """
        raise NotImplementedError

    @handler("PostToConnection")
    def post_to_connection(
        self, context: RequestContext, connection_id: _string, data: Data
    ) -> None:
        """Sends the provided data to the specified connection.

        :param connection_id: The identifier of the connection that a specific client is using.
        :param data: The data to be sent to the client specified by its connection id.
        :raises GoneException:
        :raises LimitExceededException:
        :raises PayloadTooLargeException:
        :raises ForbiddenException:
        """
        raise NotImplementedError



_B='Properties'
_A='Name'
from typing import Dict
from localstack.services.cloudformation.deployment_utils import generate_default_name_without_stack,params_list_to_dict
from localstack.services.cloudformation.service_models import GenericBaseModel
from localstack.utils.aws import aws_stack
from localstack.utils.collections import select_attributes
from localstack.utils.objects import not_none_or
class QLDBLedger(GenericBaseModel):
	@staticmethod
	def cloudformation_type():return'AWS::QLDB::Ledger'
	def get_physical_resource_id(A,attribute=None,**B):return A.props.get(_A)
	def fetch_state(A,stack_name,resources):B=aws_stack.connect_to_service('qldb');C=A.resolve_refs_recursively(stack_name,A.props.get(_A),resources);D=B.describe_ledger(Name=C);return D
	@staticmethod
	def add_defaults(resource,stack_name):
		A=resource;B=A[_B]
		if not B.get(_A):B[_A]=generate_default_name_without_stack(A['LogicalResourceId'])
	@staticmethod
	def get_deploy_templates():
		A='function'
		def B(resource_id,resources,*D,**E):A=aws_stack.connect_to_service('qldb');C=resources[resource_id][_B];B=C.get(_A);A.update_ledger(Name=B,DeletionProtection=False);A.delete_ledger(Name=B)
		def C(params,*E,**F):
			B='Tags';A=params;C=select_attributes(A,['DeletionProtection','KmsKey',_A,'PermissionsMode'])
			if A.get(B):D=params_list_to_dict(B)(A);C[B]={A:str(not_none_or(B,''))for(A,B)in D.items()}
			return C
		return{'create':{A:'create_ledger','parameters':C},'delete':{A:B}}


from typing import Dict
from localstack.aws.api import RequestContext
from localstack.aws.chain import Handler,HandlerChain
from localstack.http import Response
from localstack.services.plugins import SERIALIZERS,StateSerializer
from localstack_ext.utils.cloud_pods import CIRunManager,SyncPodStateManager
class PodStateSyncHandler(Handler):
	def __call__(A,chain,request,response):SyncPodStateManager.get().sync_pod_state()
class CIMetricPublisher(Handler):
	def __call__(A,chain,request,response):CIRunManager.get().publish_metrics()
class StateSerializerHandler(Handler):
	serializers:Dict[str,StateSerializer]
	def __init__(A):A.serializers=SERIALIZERS
	def __call__(C,chain,context,response):
		A=context
		if not A.service:return
		D=A.service.service_name;B=C.serializers.get(D)
		if not B:return
		B.update_state(B.get_context(),A.request)


from __future__ import annotations
_A=None
import logging,os,sqlite3
from shutil import copyfile
from typing import Any,Dict,List,Optional,Set,Tuple
from localstack.utils.files import mkdir
from localstack.utils.strings import long_uid
from localstack_ext.bootstrap.state_merge import merge_3way
LOG=logging.getLogger(__name__)
class DDBAttribute:
	_PRIMARY_KEY='PRIMARY KEY';_NOT_NULLABLE='NOT NULL';_DEFAULT='DEFAULT';_NULL='NULL';_TEXT_TYPES={'TEXT'};_NUMERIC_TYPES={'NUMERIC','BOOLEAN','DATE','DATETIME'};_INTEGER_TYPES={'INT','INTEGER'};_REAL_TYPES={'REAL','DOUBLE','FLOAT'};_BLOB_TYPES={'BLOB'}
	def __init__(A,name,typ,nullable,default,primary):A.name=name;A.typ=typ;A.nullable=nullable;A.default=default;A.primary=primary
	@classmethod
	def from_table_info(B,table_info):A=table_info;C=str(A[1]);D=str(A[2]);E=not bool(A[3]);F=A[4];G=bool(A[5]);return B(C,D,E,F,G)
	def get_schema(A):
		B=[A.name,A.typ]
		if A.primary:B.append(A._PRIMARY_KEY)
		if not A.nullable:B.append(A._NOT_NULLABLE)
		if A.default:B.extend([A._DEFAULT,A.default])
		return ' '.join(B).strip()
	def __str__(A):return A.get_schema()
	def _default_of_typ(A,typ_name):
		B=typ_name
		if B in A._TEXT_TYPES:return str()
		elif B in A._BLOB_TYPES:return bytes()
		elif B in{*A._NUMERIC_TYPES,*A._INTEGER_TYPES}:return int()
		elif B in A._REAL_TYPES:return float()
		else:LOG.warning(f"No type affinities found for type {B}");return _A
	def _try_cast_to_type_value(C,typ,value,def_value):
		B=def_value;A=value
		if A:
			try:
				if isinstance(A,str)and isinstance(B,bytes):return A
				else:return typ(A)
			except Exception as D:LOG.warning(f"Cannot update value type of '{A}' '{typ}' to be a {C.typ}, hence replaced it with default {B}, error: {D}")
		return B
	def _to_compatible_type_value(A,value):
		B=value
		if not B and A.nullable:return B
		if not B and A.default:return A.default
		C=A._default_of_typ(A.typ);D=type(C);E=type(B)
		if D==E:return B
		else:return A._try_cast_to_type_value(D,B,C)
	def adapt_value(A,value):
		B=A._to_compatible_type_value(value)
		if B:return B
		if A.default:return A.default
		if not A.nullable and not A.default:return A._default_of_typ(A.typ)
		return A._NULL
class DDBTableSchema:
	def __init__(A,table_name,attributes_by_name,og_schema):A.table_name=table_name;A._attributes_by_name=attributes_by_name;A._og_schema=og_schema
	@classmethod
	def from_database(B,table_name,cur):A=table_name;C=DDBTableSchema._load_fields(A,cur);D=DDBTableSchema._load_og_schema(A,cur);return B(A,C,D)
	@staticmethod
	def _sqlite_table_info(table_name):return f'PRAGMA table_info("{table_name}");'
	@staticmethod
	def _sqlite_table_schema(table_name):return f'SELECT sql FROM sqlite_master WHERE tbl_name="{table_name}" LIMIT 1;'
	@staticmethod
	def _load_fields(table_name,cur):
		C=DDBTableSchema._sqlite_table_info(table_name);cur.execute(C);D=cur.fetchall();A={}
		for E in D:B=DDBAttribute.from_table_info(E);A[B.name]=B
		return A
	@staticmethod
	def _load_og_schema(table_name,cur):A=DDBTableSchema._sqlite_table_schema(table_name);cur.execute(A);B=cur.fetchall();C=B[0][0];return C
	def get_attributes(A):return A._attributes_by_name
	def get_attribute_names(B):
		A=[]
		for C in B._attributes_by_name.values():A.append(C.name)
		return A
	def get_schema(A):B=f'CREATE TABLE "{A.table_name}"';C=','.join([B.get_schema()for B in A._attributes_by_name.values()]);D=f"{B} ({C});";return D
	def get_insert_statement(A):B=','.join(A._attributes_by_name.keys());C=','.join(['?']*len(A._attributes_by_name));D=f'INSERT INTO "{A.table_name}" ({B}) VALUES ({C});';return D
	def __str__(A):return A.get_schema()
class DDBField:
	def __init__(A,value):A.value=value
class DDBRecord:
	def __init__(A,schema,record):B=schema;A.fields_by_name=A._load_fields_by_name(B,record);A.primary_id=A._derive_primary_id(B,A.fields_by_name)
	@staticmethod
	def _derive_primary_id(schema,field_by_name):A=schema.get_attributes().values();B=[B.name for B in A if B.primary]or[B.name for B in A];C=[str(field_by_name[A].value)for A in B];D='-'.join(C);return D
	@staticmethod
	def _load_fields_by_name(schema,record):
		A={};B=schema.get_attributes()
		for (C,D) in enumerate(B.keys()):E=record[C];F=DDBField(E);A[D]=F
		return A
	def get_insert_tuple(C,schema):
		A=[];D=schema.get_attributes()
		for (E,F) in D.items():B=C.fields_by_name.get(E,_A);G=B.value if B else _A;H=F.adapt_value(G);A.append(H)
		return tuple(A)
class DDBTable:
	def __init__(A,table_name,schema):A._table_name=table_name;A._schema=schema;A._records_by_primary_id={}
	@classmethod
	def from_database(B,table_name,cur):A=table_name;C=DDBTableSchema.from_database(A,cur);return B(A,C)
	@staticmethod
	def _sql_select_all(table_name):return f'SELECT * FROM "{table_name}";'
	def offload_records(A):A._records_by_primary_id.clear()
	def load_records(A,cur):
		A.offload_records();C=A._sql_select_all(A._table_name);cur.execute(C);D=cur.fetchall()
		for E in D:B=DDBRecord(A._schema,E);A._records_by_primary_id[B.primary_id]=B
	def get_table_name(A):return A._table_name
	def set_table_name(A,table_name):B=table_name;A._table_name=B;A._schema.table_name=B
	def get_table_schema(A):return A._schema
	def dump_records(A,cur):
		if not A._records_by_primary_id:return
		C=A._schema.get_insert_statement()
		for B in A._records_by_primary_id.values():
			D=B.get_insert_tuple(A._schema)
			try:cur.execute(C,D)
			except Exception as E:LOG.warning(f"Could not insert record {B} after merge: {E}")
class DDBDatabase:
	_SQLITE_GET_TABLES="SELECT name FROM sqlite_master WHERE type='table';"
	def __init__(A,file_path):A._con=sqlite3.connect(file_path);A._cur=A._con.cursor();A._tables_by_name=A._find_tables()
	def _find_tables(A):
		A._cur.execute(A._SQLITE_GET_TABLES);D=A._cur.fetchall();B={}
		for E in D:C=E[0];B[C]=DDBTable.from_database(C,A._cur)
		return B
	def close(A):A._cur.close();A._con.close()
	def commit(A):A._con.commit()
	def commit_and_close(A):A.commit();A.close()
	@staticmethod
	def _generate_table_tmp_name(table):return f"{table.get_table_name()}_{long_uid().replace('-','_')}"
	@staticmethod
	def _sql_delete_table(table_name):return f'DROP TABLE IF EXISTS "{table_name}";'
	@staticmethod
	def _sql_rename_table(table_name,new_table_name):return f'ALTER TABLE "{table_name}" RENAME TO "{new_table_name}";'
	@staticmethod
	def _sql_pragma_enable(enabled):A='TRUE'if enabled else'FALSE';return f"PRAGMA ignore_check_constraints={A};"
	def _pragma_enable(A):B=A._sql_pragma_enable(False);A._cur.execute(B)
	def _pragma_disable(A):B=A._sql_pragma_enable(True);A._cur.execute(B)
	def _create_table(A,table):B=table.get_table_schema().get_schema();A._cur.execute(B)
	def _delete_table(A,table_name):B=A._sql_delete_table(table_name);A._cur.execute(B)
	def _rename_table(A,table_name,new_table_name):B=A._sql_rename_table(table_name,new_table_name);A._cur.execute(B)
	def load_records(A):
		for B in A._tables_by_name.values():B.load_records(A._cur)
	def offload_records(A):
		for B in A._tables_by_name.values():B.offload_records()
	def merge(A,inject,ancestor):
		C=inject;B=ancestor;A.load_records();C.load_records()
		if B:B.load_records()
		E={*A._tables_by_name.keys()};F=C._tables_by_name;G=B._tables_by_name if B else{};merge_3way(A._tables_by_name,F,G);A._pragma_disable()
		for H in E:A._delete_table(H)
		for D in A._tables_by_name.values():A._create_table(D);D.dump_records(A._cur)
		A._pragma_enable()
def merge_dynamodb(path_current,path_inject,path_ancestor=_A):
	C=path_inject;B=path_current;from localstack_ext.services.dynamodb.provider import restart_dynamodb as F;mkdir(B);G=os.listdir(B)
	for A in os.listdir(C):
		D=os.path.join(B,A);E=os.path.join(C,A)
		if A in G:merge_sqlite_dbs(D,E)
		else:copyfile(E,D);LOG.debug('Copied state from previously non-existing region file %s',A)
	F()
def merge_sqlite_dbs(sqlite_file_current,sqlite_file_inject,sqlite_file_ancestor=_A):
	D=sqlite_file_inject;C=sqlite_file_current;A=sqlite_file_ancestor;E=DDBDatabase(C);F=DDBDatabase(D);B=DDBDatabase(A)if A else _A;E.merge(F,B);E.commit_and_close();F.close()
	if B:B.close()
	LOG.debug(f"Successfully merged db {C} with {D} and {A}")


_P='DistributionConfig'
_O='OriginRequestPolicy'
_N='delete'
_M='CloudFrontOriginAccessIdentityConfig'
_L='Aliases'
_K='parameters'
_J='CallerReference'
_I='Origins'
_H='create'
_G='Comment'
_F='function'
_E='cloudfront'
_D='Id'
_C=None
_B='Name'
_A='Items'
from localstack.services.cloudformation.service_models import REF_ID_ATTRS,GenericBaseModel
from localstack.utils.aws import aws_stack
from localstack.utils.common import select_attributes,short_uid,to_bytes
from localstack_ext.utils.aws import aws_utils
class CloudFrontOriginAccessIdentity(GenericBaseModel):
	@staticmethod
	def cloudformation_type():return'AWS::CloudFront::CloudFrontOriginAccessIdentity'
	def get_physical_resource_id(A,attribute=_C,**B):return A.props.get(_D)
	def fetch_state(B,stack_name,resources):C=aws_stack.connect_to_service(_E);A=C.list_cloud_front_origin_access_identities();A=A.get('CloudFrontOriginAccessIdentityList',{}).get(_A,[]);D=B.props.get(_M,{});A=[B for B in A if B[_G]==D.get(_G)];return A and A[0]or _C
	@staticmethod
	def get_deploy_templates():
		def A(params,**C):A=params;B=A.get(_M);B[_J]=short_uid();return A
		return{_H:{_F:'create_cloud_front_origin_access_identity',_K:A}}
class CloudFrontFunction(GenericBaseModel):
	@staticmethod
	def cloudformation_type():return'AWS::CloudFront::Function'
	def fetch_state(B,stack_name,resources):C=aws_stack.connect_to_service(_E);A=C.list_functions();A=A.get('FunctionList',{}).get(_A,[]);D=B.props;A=[B for B in A if B[_B]==D.get(_B)];return A and A[0]or _C
	def get_physical_resource_id(A,attribute,**B):return A.props.get(_B)
	def get_cfn_attribute(A,attribute_name):
		B=attribute_name
		if B=='FunctionARN':return aws_utils.get_cloudfront_function_arn(A.props.get(_B))
		return super(CloudFrontFunction,A).get_cfn_attribute(B)
	@classmethod
	def get_deploy_templates(C):
		def A(params,**C):
			B='FunctionCode';A=select_attributes(params,[_B,B,'FunctionConfig'])
			if A.get(B):A[B]=to_bytes(A[B])
			return A
		def B(resource_id,resources,resource_type,func,stack_name,*F):D=C(resources[resource_id]);A=aws_stack.connect_to_service(_E);B=D.props.get(_B);E=A.describe_function(Name=B).get('ETag');return A.delete_function(Name=B,IfMatch=E)
		return{_H:{_F:'create_function',_K:A},_N:{_F:B}}
class CloudFrontOriginRequestPolicy(GenericBaseModel):
	@staticmethod
	def cloudformation_type():return'AWS::CloudFront::OriginRequestPolicy'
	def fetch_state(C,stack_name,resources):D='OriginRequestPolicyConfig';E=aws_stack.connect_to_service(_E);A=E.list_origin_request_policies();A=A.get('OriginRequestPolicyList',{}).get(_A,[]);A=[B.get(_O,{})for B in A];A=[{_D:B[_D],**B.get(D,{})}for B in A];B=C.props.get(D,{}).get(_B);B=C.resolve_refs_recursively(stack_name,B,resources);A=[C for C in A if C.get(_B)==B];return A and A[0]or _C
	def get_physical_resource_id(A,attribute,**B):return A.props.get(_D)
	@classmethod
	def get_deploy_templates(C):
		def A(resource_id,resources,resource_type,func,stack_name,*G):A=aws_stack.connect_to_service(_E);D=C(resources[resource_id]);E=D.props;B=E.get(_O).get(_D);F=A.get_origin_request_policy(Id=B).get('ETag');return A.delete_origin_request_policy(Id=B,IfMatch=F)
		return{_H:{_F:'create_origin_request_policy'},_N:{_F:A}}
class CloudFrontDistribution(GenericBaseModel):
	@staticmethod
	def cloudformation_type():return'AWS::CloudFront::Distribution'
	def get_physical_resource_id(B,attribute,**C):
		A=B.props.get(_D)
		if attribute in REF_ID_ATTRS:return A
		return aws_utils.get_cloudfront_distribution_arn(A)
	def fetch_state(F,stack_name,resources):
		E='DomainName';G=aws_stack.connect_to_service(_E);H=G.list_distributions()['DistributionList'].get(_A,[]);D=F.props[_P];A=D.get(_I,[]);A=A if isinstance(A,list)else A.get(_A);I=set([B[E]for B in A]);B=D.get(_L,[]);B=B if isinstance(B,list)else B.get(_A);J=set(B)
		for C in H:
			A=set([A.get(E)for A in C.get(_I,{})[_A]]);B=set(C.get(_L,{}).get(_A,[]))
			if I==A and J==B:return C
	@staticmethod
	def get_deploy_templates():
		def A(params,**m):
			k='Cookies';j='HTTPSPort';i='HTTPPort';h='OriginSslProtocols';g='CacheBehaviors';f='Tags';e='IamCertificateId';d='SslSupportMethod';c='AcmCertificateArn';b='QueryString';a='MinTTL';Z='TrustedSigners';Y='ResponseCode';X='CustomErrorResponses';W='ViewerCertificate';V='true';U=False;R='DefaultCacheBehavior';Q='Enabled';P='GeoRestriction';M=params;L='ForwardedValues';K='Quantity';H='CachedMethods';E='AllowedMethods';A=M[_P];A[_J]=A.get(_J)or short_uid();A[_G]=A.get(_G)or'';A['IsIPV6Enabled']=A.pop('IPV6Enabled',U);A[Q]=str(A.get(Q)).lower()==V;C=A[W]=A.get(W,{});I=A.get(X,[]);I=I.get(_A,[])if isinstance(I,dict)else I
			for S in I:S[Y]=str(S.get(Y))
			F=A[R];F[Z]=F.get(Z)or{K:0,Q:U};F[a]=F.get(a)or 600;T=F[L]=F.get(L,{});T[b]=str(T.get(b)).lower()==V
			if c in C:C['ACMCertificateArn']=C.pop(c)
			if d in C:C['SSLSupportMethod']=C.pop(d)
			if e in C:C['IAMCertificateId']=C.pop(e)
			M[f]={_A:M.get(f)or[]}
			def B(key,parent=_C):
				B=parent;B=A if B is _C else B;C=B.get(key)or[]
				if isinstance(C,dict)and _A in C and K in C:return
				B[key]={K:len(C),_A:C}
			J=A.get(g,[]);J=J.get(_A,[])if isinstance(J,dict)else J
			for N in J:B(E,N);N[E][H]=N.pop(H,[]);B(H,N[E])
			for l in A[_I]:
				D=(l or{}).get('CustomOriginConfig')
				if not D:continue
				D[h]=D.pop('OriginSSLProtocols',[]);D[i]=D.get(i,80);D[j]=D.get(j,443);B(h,D)
			B(_L);B(_I);B('OriginGroups');B(X);B(g);B(E,F);G=A.get(R,{});G[E]=G.get(E);G[E][H]=A.get(R,{}).pop(H,[]);B(H,G[E])
			if G.get(L)and not G[L].get(k):G[L][k]={'Forward':'All'}
			O=A.get('Restrictions',{})
			if O.get(P)and K not in O[P]:O[P][K]=0;O[P][_A]=[]
			return M
		return{_H:{_F:'create_distribution_with_tags',_K:{'DistributionConfigWithTags':A}}}


_D='/user/signin'
_C='username'
_B=None
_A='token'
import getpass,json,logging,os,sys,time
from typing import Any,Dict,Optional
from jose import JWTError,jwt
from localstack.config import dirs
from localstack.constants import API_ENDPOINT
from localstack.utils.common import FileMappedDocument,call_safe,safe_requests,to_str
LOG=logging.getLogger(__name__)
AUTH_CACHE_FILE='auth.json'
class AuthToken:
	def __init__(A,token,metadata=_B):A.token=token;A.metadata=metadata
	def extract_jwt(B):A=B.token.strip().split(' ')[-1];jwt.get_unverified_claims(A);return A
	def to_dict(A):return{**(A.metadata or{}),_A:A.token}
class AuthClient:
	TOKEN_REFRESH_LEAD_TIME=30
	def get_auth_token(C,username,password,headers=_B):
		D={_C:username,'password':password};A=safe_requests.post(C._api_url(_D),json.dumps(D),headers=headers)
		if not A.ok:return _B
		try:B=json.loads(to_str(A.content or'{}'));return AuthToken(token=B[_A],metadata=B)
		except Exception:pass
	def get_token_expiry(B,token):
		try:A=jwt.get_unverified_claims(token.extract_jwt());return A.get('exp')
		except JWTError:return
	def refresh_token(A,token):
		B=token;D=A.get_token_expiry(B)
		if not D or time.time()<D-A.TOKEN_REFRESH_LEAD_TIME:return B
		F=B.to_dict();C=safe_requests.put(A._api_url(_D),json.dumps(F))
		if not C.ok:raise Exception(f"Unable to obtain auth token (code {C.status_code}) - please log in again.")
		try:G=json.loads(to_str(C.content or'{}'));E=G[_A];return AuthToken(token=E[_A],metadata=E)
		except Exception as H:raise Exception(f"Unable to obtain token ({H}) - please log in again.")
	def read_credentials(C,username):
		A=username;print('Please provide your login credentials below')
		if not A:sys.stdout.write('Username: ');sys.stdout.flush();A=input()
		B=getpass.getpass();return A,B,{}
	def _api_url(A,path):return f"{API_ENDPOINT}{path}"
def get_auth_cache():return FileMappedDocument(os.path.join(dirs.cache,AUTH_CACHE_FILE),mode=384)
def login(username=_B):
	A=username;B=AuthClient();A,E,F=B.read_credentials(A);print('Verifying credentials ... (this may take a few moments)');C=B.get_auth_token(A,E,F)
	if not C:raise Exception('Unable to verify login credentials - please try again')
	D=get_auth_cache();D.update({_C:A,_A:C.token});call_safe(D.save,exception_message='error saving authentication information')
def logout():A=get_auth_cache();A.clear();A.save()


_A=None
import json,logging,os,zipfile
from typing import Dict,List,Union
from localstack.utils.json import FileMappedDocument
from localstack_ext.bootstrap.auth import get_auth_cache
from localstack_ext.bootstrap.pods.constants import ASSETS_ROOT,COMMIT_FILE,COMPRESSION_FORMAT,DELTA_LOG_DIR,HEAD_FILE,KNOWN_VER_FILE,MAX_VER_FILE,META_ZIP,METAMODELS_FILE,OBJ_STORE_DIR,PODS_CONFIG,REFS_DIR,REMOTE_FILE,REV_SUB_DIR,STATE_ZIP,VER_LOG_FILE,VER_LOG_STRUCTURE,VER_SUB_DIR,VERSION_SPACE_DIRS,VERSION_SPACE_FILES
LOG=logging.getLogger(__name__)
class PodsConfigContext:
	def __init__(A,pod_root_dir):B=pod_root_dir;A.cloud_pods_root_dir=B;A.pod_root_dir=B;A.user=_A
	def get_pod_context(A):return os.path.basename(A.pod_root_dir)
	def get_context_user(A):return A.user
	def get_pod_root_dir(A):return A.pod_root_dir
	def get_cloud_pods_root_dir(A):return A.cloud_pods_root_dir
	def get_head_path(A):return os.path.join(A.pod_root_dir,HEAD_FILE)
	def get_max_ver_path(A):return os.path.join(A.pod_root_dir,MAX_VER_FILE)
	def get_known_ver_path(A):return os.path.join(A.pod_root_dir,KNOWN_VER_FILE)
	def get_ver_log_path(A):return os.path.join(A.pod_root_dir,VER_LOG_FILE)
	def get_obj_store_path(A):return os.path.join(A.pod_root_dir,OBJ_STORE_DIR)
	def get_rev_obj_store_path(A):return os.path.join(A.get_obj_store_path(),REV_SUB_DIR)
	def get_ver_obj_store_path(A):return os.path.join(A.get_obj_store_path(),VER_SUB_DIR)
	def get_ver_refs_path(A):return os.path.join(A.pod_root_dir,REFS_DIR,VER_SUB_DIR)
	def get_rev_refs_path(A):return os.path.join(A.pod_root_dir,REFS_DIR,REV_SUB_DIR)
	def get_version_ref_file_path(A,version_ref):return os.path.join(A.get_ver_refs_path(),version_ref)
	def get_delta_log_path(A):return os.path.join(A.pod_root_dir,A.get_obj_store_path(),DELTA_LOG_DIR)
	def get_assets_root_path(A):return os.path.join(A.pod_root_dir,ASSETS_ROOT)
	def get_version_meta_archive_path(B,version,with_format=True):
		A=os.path.join(B.get_pod_root_dir(),META_ZIP.format(version_no=version))
		if not with_format:return A
		return f"{A}.{COMPRESSION_FORMAT}"
	def get_version_state_archive_path(B,version,with_format=True):
		A=os.path.join(B.get_pod_root_dir(),STATE_ZIP.format(version_no=version))
		if not with_format:return A
		return f"{A}.{COMPRESSION_FORMAT}"
	def update_ver_log(A,author,ver_no,rev_id,rev_no):
		with open(A.get_ver_log_path(),'a')as B:B.write(f"{VER_LOG_STRUCTURE.format(author=author,ver_no=ver_no,rev_rid_no=f'{rev_id}_{rev_no}')}\n")
	def create_version_symlink(A,name,key=_A):return A._create_symlink(name,key,A.get_ver_refs_path())
	def create_revision_symlink(A,name,key=_A):return A._create_symlink(name,key,A.get_rev_refs_path())
	def is_initialized(A):return A.pod_root_dir and os.path.isdir(A.pod_root_dir)
	def _create_symlink(A,name,key,path):
		B=os.path.relpath(path,start=A.get_pod_root_dir());C=os.path.join(B,name)
		if key:
			D=os.path.join(path,name)
			with open(D,'w')as E:E.write(key)
		return C
	def get_head_key(A):return A._get_key(A.get_head_path())
	def get_max_ver_key(A):return A._get_key(A.get_max_ver_path())
	def _get_key(B,path):
		with open(path,'r')as A:C=A.readline().strip()
		D=B.get_pod_absolute_path(C)
		with open(D,'r')as A:E=A.readline();return E
	def get_pod_absolute_path(A,rel_path):return os.path.join(A.get_pod_root_dir(),rel_path)
	def get_obj_file_path(A,key):return os.path.join(A.get_obj_store_path(),key)
	def get_remote_info_path(A):return os.path.join(A.pod_root_dir,REMOTE_FILE)
	def is_remotely_managed(A,pod_name=_A):
		B=pod_name
		if B:return os.path.isfile(os.path.join(A.cloud_pods_root_dir,B,REMOTE_FILE))
		else:return os.path.isfile(A.get_remote_info_path())
	def set_pod_context(A,pod_name):B=get_auth_cache();C=B.get('username','unknown');A.pod_root_dir=os.path.join(A.cloud_pods_root_dir,pod_name);A.user=C
	def pod_exists_locally(A,pod_name):return os.path.isdir(os.path.join(A.cloud_pods_root_dir,pod_name))
	def rename_pod(A,new_pod_name):C=A.get_pod_root_dir();B=os.path.join(A.cloud_pods_root_dir,new_pod_name);os.rename(C,B);A.set_pod_context(B)
	def get_pod_name(A):return os.path.basename(A.get_pod_root_dir())
	def get_version_space_dir_paths(A):return[os.path.join(A.get_pod_root_dir(),B)for B in VERSION_SPACE_DIRS]
	def get_version_space_file_paths(A):return[os.path.join(A.get_pod_root_dir(),B)for B in VERSION_SPACE_FILES]
	def get_pods_config_cache(A,conf_cache_file=PODS_CONFIG):return FileMappedDocument(os.path.join(A.cloud_pods_root_dir,conf_cache_file),mode=384)
	def save_pods_config(B,options):A=B.get_pods_config_cache();A.update(options);A.save()
	def metamodel_file(D,revision,version=_A,absolute=False):
		B=version;A=revision
		if not A:return METAMODELS_FILE
		C=f"metamodel_commit_{A}.json"
		if absolute:
			if B is _A:raise Exception('Missing pod version when constructing revision metamodel file path')
			C=os.path.join(D.metadata_dir(B))
		return C
	@staticmethod
	def commit_metamodel_file(commit_no):return COMMIT_FILE.format(commit_no=commit_no)
	def metadata_dir(A,version):return os.path.join(A.get_pod_root_dir(),META_ZIP.format(version_no=version))
	def get_version_meta_archive(B,version):
		A=B.get_version_meta_archive_path(version)
		if os.path.isfile(A):return A
	def get_version_state_archive(B,version):
		A=B.get_version_state_archive_path(version)
		if os.path.isfile(A):return A
def zip_directories(zip_dest,directories):
	B=zip_dest;from localstack.utils.testutil import create_zip_file_python as C
	for A in directories:C(source_path=A,content_root=os.path.basename(A),base_dir=A,zip_file=B,mode='a')
	return B
def add_file_to_archive(archive,entry_name,content):
	with zipfile.ZipFile(archive,'a')as A:A.writestr(entry_name,content)
def read_file_from_archive(archive_path,file_name):
	B=file_name;A=archive_path
	try:
		with zipfile.ZipFile(A)as C:D=json.loads(C.read(B));return json.dumps(D)
	except Exception as E:LOG.debug(f"Could not find {B} in archive {A}: {E}")


import sys
from typing import List, Optional

if sys.version_info >= (3, 8):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict

from localstack.aws.api import RequestContext, ServiceException, ServiceRequest, handler

ErrorCode = str
ErrorMessage = str
IonText = str
LedgerName = str
PageToken = str
SessionToken = str
Statement = str
TransactionId = str


class BadRequestException(ServiceException):
    """Returned if the request is malformed or contains an error such as an
    invalid parameter value or a missing required parameter.
    """

    Message: Optional[ErrorMessage]
    Code: Optional[ErrorCode]


class CapacityExceededException(ServiceException):
    """Returned when the request exceeds the processing capacity of the ledger."""

    Message: Optional[ErrorMessage]


class InvalidSessionException(ServiceException):
    """Returned if the session doesn't exist anymore because it timed out or
    expired.
    """

    Message: Optional[ErrorMessage]
    Code: Optional[ErrorCode]


class LimitExceededException(ServiceException):
    """Returned if a resource limit such as number of active sessions is
    exceeded.
    """

    Message: Optional[ErrorMessage]


class OccConflictException(ServiceException):
    """Returned when a transaction cannot be written to the journal due to a
    failure in the verification phase of *optimistic concurrency control*
    (OCC).
    """

    Message: Optional[ErrorMessage]


class RateExceededException(ServiceException):
    """Returned when the rate of requests exceeds the allowed throughput."""

    Message: Optional[ErrorMessage]


class AbortTransactionRequest(TypedDict, total=False):
    """Contains the details of the transaction to abort."""


ProcessingTimeMilliseconds = int


class TimingInformation(TypedDict, total=False):
    """Contains server-side performance information for a command. Amazon QLDB
    captures timing information between the times when it receives the
    request and when it sends the corresponding response.
    """

    ProcessingTimeMilliseconds: Optional[ProcessingTimeMilliseconds]


class AbortTransactionResult(TypedDict, total=False):
    """Contains the details of the aborted transaction."""

    TimingInformation: Optional[TimingInformation]


CommitDigest = bytes


class CommitTransactionRequest(TypedDict, total=False):
    """Contains the details of the transaction to commit."""

    TransactionId: TransactionId
    CommitDigest: CommitDigest


WriteIOs = int
ReadIOs = int


class IOUsage(TypedDict, total=False):
    """Contains I/O usage metrics for a command that was invoked."""

    ReadIOs: Optional[ReadIOs]
    WriteIOs: Optional[WriteIOs]


class CommitTransactionResult(TypedDict, total=False):
    """Contains the details of the committed transaction."""

    TransactionId: Optional[TransactionId]
    CommitDigest: Optional[CommitDigest]
    TimingInformation: Optional[TimingInformation]
    ConsumedIOs: Optional[IOUsage]


class EndSessionRequest(TypedDict, total=False):
    """Specifies a request to end the session."""


class EndSessionResult(TypedDict, total=False):
    """Contains the details of the ended session."""

    TimingInformation: Optional[TimingInformation]


IonBinary = bytes


class ValueHolder(TypedDict, total=False):
    """A structure that can contain a value in multiple encoding formats."""

    IonBinary: Optional[IonBinary]
    IonText: Optional[IonText]


StatementParameters = List[ValueHolder]


class ExecuteStatementRequest(TypedDict, total=False):
    """Specifies a request to execute a statement."""

    TransactionId: TransactionId
    Statement: Statement
    Parameters: Optional[StatementParameters]


ValueHolders = List[ValueHolder]


class Page(TypedDict, total=False):
    """Contains details of the fetched page."""

    Values: Optional[ValueHolders]
    NextPageToken: Optional[PageToken]


class ExecuteStatementResult(TypedDict, total=False):
    """Contains the details of the executed statement."""

    FirstPage: Optional[Page]
    TimingInformation: Optional[TimingInformation]
    ConsumedIOs: Optional[IOUsage]


class FetchPageRequest(TypedDict, total=False):
    """Specifies the details of the page to be fetched."""

    TransactionId: TransactionId
    NextPageToken: PageToken


class FetchPageResult(TypedDict, total=False):
    """Contains the page that was fetched."""

    Page: Optional[Page]
    TimingInformation: Optional[TimingInformation]
    ConsumedIOs: Optional[IOUsage]


class StartTransactionRequest(TypedDict, total=False):
    """Specifies a request to start a transaction."""


class StartSessionRequest(TypedDict, total=False):
    """Specifies a request to start a new session."""

    LedgerName: LedgerName


class SendCommandRequest(ServiceRequest):
    SessionToken: Optional[SessionToken]
    StartSession: Optional[StartSessionRequest]
    StartTransaction: Optional[StartTransactionRequest]
    EndSession: Optional[EndSessionRequest]
    CommitTransaction: Optional[CommitTransactionRequest]
    AbortTransaction: Optional[AbortTransactionRequest]
    ExecuteStatement: Optional[ExecuteStatementRequest]
    FetchPage: Optional[FetchPageRequest]


class StartTransactionResult(TypedDict, total=False):
    """Contains the details of the started transaction."""

    TransactionId: Optional[TransactionId]
    TimingInformation: Optional[TimingInformation]


class StartSessionResult(TypedDict, total=False):
    """Contains the details of the started session."""

    SessionToken: Optional[SessionToken]
    TimingInformation: Optional[TimingInformation]


class SendCommandResult(TypedDict, total=False):
    StartSession: Optional[StartSessionResult]
    StartTransaction: Optional[StartTransactionResult]
    EndSession: Optional[EndSessionResult]
    CommitTransaction: Optional[CommitTransactionResult]
    AbortTransaction: Optional[AbortTransactionResult]
    ExecuteStatement: Optional[ExecuteStatementResult]
    FetchPage: Optional[FetchPageResult]


class QldbSessionApi:

    service = "qldb-session"
    version = "2019-07-11"

    @handler("SendCommand")
    def send_command(
        self,
        context: RequestContext,
        session_token: SessionToken = None,
        start_session: StartSessionRequest = None,
        start_transaction: StartTransactionRequest = None,
        end_session: EndSessionRequest = None,
        commit_transaction: CommitTransactionRequest = None,
        abort_transaction: AbortTransactionRequest = None,
        execute_statement: ExecuteStatementRequest = None,
        fetch_page: FetchPageRequest = None,
    ) -> SendCommandResult:
        """Sends a command to an Amazon QLDB ledger.

        Instead of interacting directly with this API, we recommend using the
        QLDB driver or the QLDB shell to execute data transactions on a ledger.

        -  If you are working with an AWS SDK, use the QLDB driver. The driver
           provides a high-level abstraction layer above this *QLDB Session*
           data plane and manages ``SendCommand`` API calls for you. For
           information and a list of supported programming languages, see
           `Getting started with the
           driver <https://docs.aws.amazon.com/qldb/latest/developerguide/getting-started-driver.html>`__
           in the *Amazon QLDB Developer Guide*.

        -  If you are working with the AWS Command Line Interface (AWS CLI), use
           the QLDB shell. The shell is a command line interface that uses the
           QLDB driver to interact with a ledger. For information, see
           `Accessing Amazon QLDB using the QLDB
           shell <https://docs.aws.amazon.com/qldb/latest/developerguide/data-shell.html>`__.

        :param session_token: Specifies the session token for the current command.
        :param start_session: Command to start a new session.
        :param start_transaction: Command to start a new transaction.
        :param end_session: Command to end the current session.
        :param commit_transaction: Command to commit the specified transaction.
        :param abort_transaction: Command to abort the current transaction.
        :param execute_statement: Command to execute a statement in the specified transaction.
        :param fetch_page: Command to fetch a page.
        :returns: SendCommandResult
        :raises BadRequestException:
        :raises InvalidSessionException:
        :raises OccConflictException:
        :raises RateExceededException:
        :raises LimitExceededException:
        :raises CapacityExceededException:
        """
        raise NotImplementedError



_k='DomainIAMRoleName'
_j='Domain'
_i='CopyTagsToSnapshot'
_h='DeletionProtection'
_g='EnableCloudwatchLogsExports'
_f='EnableIAMDatabaseAuthentication'
_e='KmsKeyId'
_d='StorageEncrypted'
_c='PreferredMaintenanceWindow'
_b='PreferredBackupWindow'
_a='OptionGroupName'
_Z='MasterUserPassword'
_Y='MasterUsername'
_X='EngineVersion'
_W='Engine'
_V='VpcSecurityGroupIds'
_U='CharacterSetName'
_T='Endpoint'
_S='Endpoint.Port'
_R='localhost'
_Q='Endpoint.Address'
_P='Properties'
_O='Family'
_N='BackupRetentionPeriod'
_M='Description'
_L='Tags'
_K='delete'
_J='create'
_I='rds'
_H='DBParameterGroupName'
_G='DBInstanceIdentifier'
_F='parameters'
_E='DBSubnetGroupName'
_D='Port'
_C='DBClusterIdentifier'
_B='function'
_A=None
from typing import Dict
from localstack.services.cloudformation.service_models import GenericBaseModel
from localstack.utils.aws import aws_stack
from localstack.utils.common import short_uid
class RDSDBSubnetGroup(GenericBaseModel):
	@staticmethod
	def cloudformation_type():return'AWS::RDS::DBSubnetGroup'
	def get_physical_resource_id(A,attribute=_A,**B):return A.props.get(_E)
	def fetch_state(A,stack_name,resources):B=aws_stack.connect_to_service(_I);C=A.resolve_refs_recursively(stack_name,A.props.get(_E),resources);D=B.describe_db_subnet_groups()['DBSubnetGroups'];E=[A for A in D if A[_E]==C];return(E or[_A])[0]
	@staticmethod
	def get_deploy_templates():return{_J:{_B:'create_db_subnet_group'},_K:{_B:'delete_db_subnet_group',_F:{_E:_E}}}
class RDSDBCluster(GenericBaseModel):
	@staticmethod
	def cloudformation_type():return'AWS::RDS::DBCluster'
	@staticmethod
	def add_defaults(resource,stack_name):
		A=resource[_P]
		if not A.get(_C):A[_C]=f"dbc-{short_uid()}"
	def get_physical_resource_id(A,attribute=_A,**B):return A.props.get(_C)
	def fetch_state(A,stack_name,resources):B=aws_stack.connect_to_service(_I);C=B.describe_db_clusters().get('DBClusters',[]);D=A.resolve_refs_recursively(stack_name,A.props.get(_C),resources);E=[A for A in C if A[_C]==D];return(E or[_A])[0]
	def get_cfn_attribute(B,attribute):
		A=attribute
		if A==_Q:return _R
		C=B.props
		if A==_S:return C.get(_D)or C.get(_T,{}).get(_D)
		return super(RDSDBCluster,B).get_cfn_attribute(A)
	@staticmethod
	def get_deploy_templates():
		E='ScalingConfiguration';F=['AvailabilityZones',_N,_U,'DatabaseName',_C,'DBClusterParameterGroupName',_V,_E,_W,_X,_D,_Y,_Z,_a,_b,_c,'ReplicationSourceIdentifier',_L,_d,_e,'PreSignedUrl',_f,'BacktrackWindow',_g,'EngineMode',E,_h,'GlobalClusterIdentifier','EnableHttpEndpoint',_i,_j,_k,'EnableGlobalWriteForwarding','SourceRegion']
		def A(params,**G):
			D='MaxCapacity';C='MinCapacity';B=params;B={A:C for(A,C)in B.items()if A in F};A=B.get(E)
			if A:
				if C in A:A[C]=int(A[C])
				if D in A:A[D]=int(A[D])
			return B
		B={_J:{_B:'create_db_cluster',_F:A,'types':{_N:int,_D:int}},_K:{_B:'delete_db_cluster',_F:[_C]}};return B
class RDSDBParameterGroup(GenericBaseModel):
	@staticmethod
	def cloudformation_type():return'AWS::RDS::DBParameterGroup'
	def get_physical_resource_id(A,attribute=_A,**B):return A.props.get(_H)
	def fetch_state(A,stack_name,resources):C=resources;B=stack_name;E=aws_stack.connect_to_service(_I);D=A.props;F=A.resolve_refs_recursively(B,D.get(_M),C);G=A.resolve_refs_recursively(B,D.get(_O),C);H=E.describe_db_parameter_groups()['DBParameterGroups'];I=[A for A in H if A[_O]==G and A[_M]==F];return(I or[_A])[0]
	@staticmethod
	def get_deploy_templates():return{_J:{_B:'create_db_parameter_group',_F:{_H:_H,'DBParameterGroupFamily':_O,_M:_M,_L:_L}},_K:{_B:'delete_db_parameter_group',_F:[_H]}}
class RDSDBInstance(GenericBaseModel):
	@staticmethod
	def cloudformation_type():return'AWS::RDS::DBInstance'
	@staticmethod
	def add_defaults(resource,stack_name):
		A=resource[_P]
		if not A.get(_G):A[_G]=f"db-{short_uid()}"
	def get_physical_resource_id(A,attribute=_A,**B):return A.props.get(_G)
	def get_cfn_attribute(B,attribute):
		A=attribute
		if A==_Q:return _R
		C=B.props
		if A==_S:return C.get(_T,{}).get(_D)or C.get(_D)
		return super(RDSDBInstance,B).get_cfn_attribute(A)
	def fetch_state(B,stack_name,resources):A=B.props.get(_G);A=B.resolve_refs_recursively(stack_name,A,resources);C=aws_stack.connect_to_service(_I);D=C.describe_db_instances()['DBInstances'];E=[B for B in D if B[_G]==A];return(E or[_A])[0]
	@staticmethod
	def get_deploy_templates():A='AllocatedStorage';B=['DBName',_G,A,'DBInstanceClass',_W,_Y,_Z,'DBSecurityGroups','AvailabilityZone',_E,_c,_H,_N,_b,_D,'MultiAZ',_X,'AutoMinorVersionUpgrade','LicenseModel','Iops',_a,_U,'NcharCharacterSetName','PubliclyAccessible',_L,_C,'StorageType','TdeCredentialArn','TdeCredentialPassword',_d,_e,_j,_i,'MonitoringInterval','MonitoringRoleArn',_k,'PromotionTier','Timezone',_f,'EnablePerformanceInsights','PerformanceInsightsKMSKeyId','PerformanceInsightsRetentionPeriod',_g,'ProcessorFeatures',_h,'MaxAllocatedStorage','EnableCustomerOwnedIp'];return{_J:{_B:'create_db_instance',_F:B+[{_V:'VPCSecurityGroups'}],'types':{A:int,_D:int}},_K:{_B:'delete_db_instance',_F:[_G]}}


_A='ClusterArn'
from localstack.services.cloudformation.service_models import GenericBaseModel
from localstack.utils.aws import aws_stack
from localstack.utils.common import clone
class KafkaCluster(GenericBaseModel):
	@staticmethod
	def cloudformation_type():return'AWS::MSK::Cluster'
	def get_physical_resource_id(A,attribute=None,**B):return A.props.get(_A)
	def fetch_state(B,stack_name,resources):C='ClusterName';D=aws_stack.connect_to_service('kafka');E=B.resolve_refs_recursively(stack_name,B.props[C],resources);A=D.list_clusters()['ClusterInfoList'];A=[B for B in A if B[C]==E];return(A or[None])[0]
	@classmethod
	def get_deploy_templates(B):
		C='parameters';A='function'
		def D(params,resource_id,resources,**F):
			D='EBSStorageInfo';E=B(resources[resource_id]);C=clone(E.props);A=C.get('BrokerNodeGroupInfo',{}).get('StorageInfo',{})
			if D in A:A['EbsStorageInfo']=A.pop(D)
			return C
		def E(params,resource_id,resources,**C):A=B(resources[resource_id]);return{_A:A.props.get(_A)}
		return{'create':{A:'create_cluster',C:D},'delete':{A:'delete_cluster',C:E}}


_C=False
_B=None
_A=True
import json,logging,os,shutil,zipfile
from typing import Dict,List,Optional,Set,Tuple,Union
from localstack import config as localstack_config
from localstack.constants import TEST_AWS_ACCOUNT_ID
from localstack.utils.common import cp_r,mkdir,rm_rf,save_file,short_uid,to_str
from localstack.utils.http import safe_requests
from localstack_ext.bootstrap.pods.constants import CLOUD_PODS_DIR,COMPRESSION_FORMAT,DEFAULT_POD_DIR,NIL_PTR,STATE_ZIP,VER_SYMLINK,VERSION_SERVICE_INFO_FILE,VERSION_SPACE_ARCHIVE
from localstack_ext.bootstrap.pods.models import Commit,Revision,Serialization,StateFileRef,Version
from localstack_ext.bootstrap.pods.object_storage import get_object_storage_provider
from localstack_ext.bootstrap.pods.utils.common import PodsConfigContext,read_file_from_archive,zip_directories
from localstack_ext.bootstrap.pods.utils.hash_utils import compute_file_hash,compute_revision_hash,compute_version_archive_hash,random_hash
from localstack_ext.bootstrap.pods.utils.merge_utils import CloudPodsMergeManager,create_tmp_archives_by_serialization_mechanism
from localstack_ext.bootstrap.pods.utils.metamodel_utils import CommitMetamodelUtils,MetamodelDelta,MetamodelDeltaMethod
from localstack_ext.bootstrap.pods.utils.remote_utils import extract_meta_and_state_archives,merge_version_space,register_remote
from localstack_ext.bootstrap.state_utils import API_STATES_DIR,DYNAMODB_DIR,KINESIS_DIR,persist_object
LOG=logging.getLogger(__name__)
ROOT_DIR_LOOKUP={str(Serialization.KINESIS):KINESIS_DIR,str(Serialization.DDB):DYNAMODB_DIR,str(Serialization.MAIN):API_STATES_DIR}
class PodsApi:
	def __init__(A):
		B=os.environ.get('POD_DIR')
		if not B:B=os.path.join(localstack_config.dirs.tmp,DEFAULT_POD_DIR)
		B=os.path.join(B,CLOUD_PODS_DIR);A.config_context=PodsConfigContext(B);A.object_storage=get_object_storage_provider(A.config_context);A.commit_metamodel_utils=CommitMetamodelUtils(A.config_context)
	def init(A,pod_name='My-Pod'):
		C=pod_name
		if A.config_context.pod_exists_locally(pod_name=C):LOG.warning(f"Pod with name {C} already exists locally");return
		A.config_context.set_pod_context(C)
		def H():mkdir(A.config_context.get_pod_root_dir());mkdir(A.config_context.get_ver_refs_path());mkdir(A.config_context.get_rev_refs_path());mkdir(A.config_context.get_ver_obj_store_path());mkdir(A.config_context.get_rev_obj_store_path());mkdir(A.config_context.get_delta_log_path())
		H();D=random_hash();I=random_hash();E=Revision(hash_ref=D,parent_ptr=NIL_PTR,creator=A.config_context.get_context_user(),rid=short_uid(),revision_number=0,state_files=set());F=Version(hash_ref=I,parent_ptr=NIL_PTR,creator=A.config_context.get_context_user(),comment='Init version',active_revision_ptr=D,outgoing_revision_ptrs={D},incoming_revision_ptr=_B,state_files=set(),version_number=0);K,J=A.object_storage.upsert_objects(E,F);G=A.config_context.create_version_symlink(VER_SYMLINK.format(ver_no=F.version_number),J)
		with open(A.config_context.get_head_path(),'w')as B:B.write(G)
		with open(A.config_context.get_max_ver_path(),'w')as B:B.write(G)
		with open(A.config_context.get_known_ver_path(),'w')as B:B.write(G)
		A.config_context.update_ver_log(author=A.config_context.get_context_user(),ver_no=F.version_number,rev_id=E.rid,rev_no=E.revision_number);LOG.debug(f"Successfully initiated CloudPods for pod at {A.config_context.get_pod_root_dir()}")
	def init_remote(A,version_space_archive,meta_archives,state_archives,remote_info,pod_name):
		B=version_space_archive;A.config_context.set_pod_context(pod_name=pod_name);mkdir(A.config_context.get_pod_root_dir())
		with zipfile.ZipFile(B)as C:C.extractall(A.config_context.get_pod_root_dir());LOG.debug('Successfully extracted version space zip')
		rm_rf(B);extract_meta_and_state_archives(meta_archives=meta_archives,state_archives=state_archives,config_context=A.config_context);D=A._get_max_version();E=A.config_context.create_version_symlink(name=VER_SYMLINK.format(ver_no=D.version_number))
		with open(A.config_context.get_head_path(),'w')as F:F.write(E)
		register_remote(remote_info=remote_info,config_context=A.config_context)
	def commit(B,message=_B):
		A,F=B._get_expansion_point_with_head();D=compute_revision_hash(A,B.config_context.get_obj_store_path());G=set()
		if A.parent_ptr!=NIL_PTR:H=_B;E=B.object_storage.get_revision_by_key(A.parent_ptr);G=E.state_files;E.assoc_commit.head_ptr=D;B.object_storage.upsert_objects(E)
		else:H=F.hash_ref
		B.object_storage.update_revision_key(old_key=A.hash_ref,new_key=D,referenced_by_version=H);A.hash_ref=D;C=Revision(hash_ref=random_hash(),state_files=set(),parent_ptr=D,creator=A.creator,rid=short_uid(),revision_number=A.revision_number+1);I=B.create_delta_log(G,A.state_files);J=Commit(tail_ptr=A.hash_ref,head_ptr=C.hash_ref,message=message,delta_log_ptr=I);A.assoc_commit=J;B.object_storage.upsert_objects(C,A);B.config_context.update_ver_log(author=C.creator,ver_no=F.version_number,rev_id=C.rid,rev_no=C.revision_number);return A
	def merge_from_remote(A,version_space_archive,meta_archives,state_archives):merge_version_space(version_space_archive,config_context=A.config_context);extract_meta_and_state_archives(meta_archives=meta_archives,state_archives=state_archives,config_context=A.config_context)
	def is_remotely_managed(A):return A.config_context.is_remotely_managed()
	def rename_pod(B,new_pod_name):
		A=new_pod_name
		if B.config_context.pod_exists_locally(A):LOG.warning(f"{A} already exists locally");return _C
		B.config_context.rename_pod(A);return _A
	def list_locally_available_pods(A,show_remote_or_local=_A):
		mkdir(A.config_context.cloud_pods_root_dir);C=[B for B in os.listdir(A.config_context.cloud_pods_root_dir)if not B.endswith('.json')]
		if not show_remote_or_local:return set(C)
		D=set()
		for B in C:E=f"remote/{B}"if A.config_context.is_remotely_managed(B)else f"local/{B}";D.add(E)
		return D
	def push(A,comment=_B,include_assets=_A):
		C,F=A._get_expansion_point_with_head();G=A._get_max_version();D=Revision(hash_ref=random_hash(),state_files=set(),parent_ptr=NIL_PTR,creator=C.creator,rid=short_uid(),revision_number=0);E=G.version_number+1
		if F.version_number!=G.version_number:A.merge_expansion_point_with_max()
		else:A._create_state_directory(E,state_file_refs=C.state_files)
		B=Version(hash_ref=compute_version_archive_hash(E,A.config_context.get_version_state_archive(E)),state_files=set(),parent_ptr=G.hash_ref,creator=C.creator,comment=comment,active_revision_ptr=D.hash_ref,outgoing_revision_ptrs={D.hash_ref},incoming_revision_ptr=C.hash_ref,version_number=E)
		if C.parent_ptr!=NIL_PTR:I=A.object_storage.get_revision_by_key(C.parent_ptr);J=I.state_files;H=A.create_delta_log(J,B.state_files)
		else:H=A.create_delta_log(C.state_files,set())
		K=Commit(tail_ptr=C.hash_ref,head_ptr=B.hash_ref,message='Finalizing commit',delta_log_ptr=H);C.state_files=B.state_files;C.assoc_commit=K;F.active_revision_ptr=NIL_PTR;A.object_storage.upsert_objects(F,C,D,B);A._update_head(B.version_number,B.hash_ref);A._update_max_ver(B.version_number,B.hash_ref);A._add_known_ver(B.version_number,B.hash_ref)
		if include_assets:A._add_assets_to_version_state_archive(B.version_number)
		A.commit_metamodel_utils.create_metadata_archive(B);A.config_context.update_ver_log(author=C.creator,ver_no=B.version_number,rev_id=D.rid,rev_no=D.revision_number);return B
	def set_pod_context(A,pod_name):A.config_context.set_pod_context(pod_name)
	def create_state_file_from_fs(D,path,file_name,service,region,root,serialization,account_id=TEST_AWS_ACCOUNT_ID):
		E=file_name;B=os.path.join(path,E);C=compute_file_hash(B);A=path.split(f"{root}/")
		if len(A)>1:A=A[1]
		else:A=''
		shutil.copy(B,os.path.join(D.config_context.get_obj_store_path(),C));F=StateFileRef(hash_ref=C,rel_path=A,file_name=E,size=os.path.getsize(B),service=service,region=region,account_id=account_id,serialization=serialization);D._add_state_file_to_expansion_point(F);return C
	def _create_state_file_from_in_memory_blob(B,blob):D=random_hash();A=os.path.join(B.config_context.get_obj_store_path(),D);persist_object(blob,A);C=compute_file_hash(A);E=os.path.join(B.config_context.get_obj_store_path(),C);os.rename(A,E);return C
	def _get_state_file_path(B,key):
		A=os.path.join(B.config_context.get_obj_store_path(),key)
		if os.path.isfile(A):return A
		LOG.warning(f"No state file with found with key: {key}")
	def _add_state_file_to_expansion_point(B,state_file):C=state_file;A,E=B._get_expansion_point_with_head();D=set(filter(lambda sf:not sf.congruent(C),A.state_files));D.add(C);A.state_files=D;B.object_storage.upsert_objects(A)
	def list_state_files(B,key):
		A=B.object_storage.get_revision_or_version_by_key(key)
		if A:return A.state_files_info()
		LOG.debug(f"No Version or Revision associated to {key}")
	def get_version_info(D,version_no):
		B=version_no;C=D.config_context.get_version_meta_archive(B)
		if not C:LOG.warning(f"No Info found for version {B}");return
		A=read_file_from_archive(C,VERSION_SERVICE_INFO_FILE);A=json.loads(to_str(A or'{}'));return A
	def create_version_space_archive(A):
		B=os.path.join(A.config_context.get_pod_root_dir(),VERSION_SPACE_ARCHIVE);rm_rf(B);C=zip_directories(zip_dest=B,directories=A.config_context.get_version_space_dir_paths())
		with zipfile.ZipFile(C,'a')as E:
			for D in A.config_context.get_version_space_file_paths():E.write(D,arcname=os.path.basename(D))
		return C
	def get_head(A):return A.object_storage.get_version_by_key(A.config_context.get_head_key())
	def _get_max_version(A):return A.object_storage.get_version_by_key(A.config_context.get_max_ver_key())
	def get_max_version_no(A):
		with open(A.config_context.get_max_ver_path())as B:return int(os.path.basename(B.readline()))
	def _get_expansion_point_with_head(A):B=A.get_head();C=A.object_storage.get_revision_by_key(key=B.active_revision_ptr);D=A.object_storage.get_terminal_revision(revision_path_root=C);return D,B
	def push_overwrite(A,version,comment):
		B=version;D,F=A._get_expansion_point_with_head()
		if B>A.get_max_version_no():LOG.debug('Attempted to overwrite a non existing version.. Aborting');return _C
		C=A.get_version_by_number(B);A._create_state_directory(version_number=B,state_file_refs=D.state_files);E=A.config_context.metamodel_file(D.revision_number);A.commit_metamodel_utils.create_metadata_archive(C,overwrite=_A,metamodels_file=E);C.comment=comment;A.object_storage.upsert_objects(C);return _A
	def _add_assets_to_version_state_archive(A,version_number,cleanup=_A):
		D=A.config_context.get_version_state_archive_path(version=version_number);B=A.config_context.get_assets_root_path()
		with zipfile.ZipFile(D,'a')as E:
			for (F,I,G) in os.walk(B):
				for H in G:C=os.path.join(F,H);E.write(filename=C,arcname=os.path.relpath(C,start=A.config_context.get_pod_root_dir()))
		if cleanup:rm_rf(B)
	@staticmethod
	def _get_dst_path_for_state_file(version_state_dir,state_file):
		C=version_state_dir;A=state_file
		if A.serialization in[str(Serialization.KINESIS),str(Serialization.DDB)]:B=os.path.join(C,ROOT_DIR_LOOKUP[A.serialization])
		else:B=os.path.join(C,ROOT_DIR_LOOKUP[A.serialization],A.rel_path)
		mkdir(B);return B
	def _create_state_directory(C,version_number,state_file_refs,delete_files=_C,archive=_A):
		A=os.path.join(C.config_context.get_pod_root_dir(),STATE_ZIP.format(version_no=version_number));mkdir(A)
		for B in state_file_refs:
			try:
				E=C._get_dst_path_for_state_file(A,B);D=C.object_storage.get_state_file_location_by_key(B.hash_ref);F=os.path.join(E,B.file_name);shutil.copy(D,F)
				if delete_files:os.remove(D)
			except Exception as G:LOG.warning(f"Failed to locate state file with rel path: {B.rel_path}: {G}")
		if archive:shutil.make_archive(A,COMPRESSION_FORMAT,root_dir=A);rm_rf(A);return f"{A}.{COMPRESSION_FORMAT}"
		return A
	def set_active_version(A,version_no,commit_before=_C):
		B=version_no;C=A.load_version_references()
		for (D,E) in C:
			if D==B:
				if commit_before:A.commit()
				A._set_active_version(E);return _A
		LOG.info(f"Version with number {B} not found");return _C
	def _set_active_version(A,key):
		C=key;E=A.get_head()
		if E.hash_ref!=C and A.object_storage.version_exists(C):
			B=A.object_storage.get_version_by_key(C);A._update_head(B.version_number,C)
			if B.active_revision_ptr==NIL_PTR:D=Revision(hash_ref=random_hash(),state_files=set(),parent_ptr=NIL_PTR,creator=A.config_context.get_context_user(),rid=short_uid(),revision_number=0);B.active_revision_ptr=D.hash_ref;B.outgoing_revision_ptrs.add(D.hash_ref);A.object_storage.upsert_objects(D,B)
	def get_version_by_number(A,version_no):
		B=version_no;D=A.load_version_references();C=next((A[1]for A in D if A[0]==B),_B)
		if not C:LOG.warning(f"Could not find version number {B}");return
		return A.object_storage.get_version_by_key(C)
	def load_version_references(B):
		C={}
		with open(B.config_context.get_known_ver_path(),'r')as D:
			E=D.readlines()
			for A in E:
				A=B.config_context.get_pod_absolute_path(A.rstrip())
				with open(A,'r')as F:C[int(os.path.basename(A))]=F.readline()
		return sorted(C.items(),key=lambda x:x[0],reverse=_A)
	def list_versions(A):B=A.load_version_references();C=[A.object_storage.get_version_by_key(C).info_str()for(D,C)in B];return C
	def list_version_commits(B,version_no):
		C=version_no
		if C==-1:D=B._get_max_version()
		else:D=B.get_version_by_number(C)
		if not D:return[]
		G=[];A=B.object_storage.get_revision_by_key(D.incoming_revision_ptr)
		while A:
			H=A.assoc_commit;E=A.revision_number
			if E!=0:F=f"Revision-{E-1}"
			elif C!=0:F=f"Version-{C}"
			else:F='Empty state'
			I=f"Revision-{E}";G.append(H.info_str(from_node=F,to_node=I));A=B.object_storage.get_revision_by_key(A.parent_ptr)
		return G
	def _update_head(A,new_head_ver_no,new_head_key):
		with open(A.config_context.get_head_path(),'w')as C:B=A.config_context.create_version_symlink(VER_SYMLINK.format(ver_no=new_head_ver_no),new_head_key);C.write(B);return B
	def _update_max_ver(A,new_max_ver_no,new_max_ver_key):
		with open(A.config_context.get_max_ver_path(),'w')as C:B=A.config_context.create_version_symlink(VER_SYMLINK.format(ver_no=new_max_ver_no),new_max_ver_key);C.write(B);return B
	def _add_known_ver(A,new_ver_no,new_ver_key):
		with open(A.config_context.get_known_ver_path(),'a')as C:B=A.config_context.create_version_symlink(VER_SYMLINK.format(ver_no=new_ver_no),new_ver_key);C.write(f"\n{B}");return B
	def merge_expansion_point_with_max(A):
		L,D=A._get_expansion_point_with_head();E=A.get_max_version_no();M=E+1;F=A._create_state_directory(version_number=M,state_file_refs=L.state_files);N=A.config_context.get_version_state_archive(E);G=create_tmp_archives_by_serialization_mechanism(F);O=create_tmp_archives_by_serialization_mechanism(N)
		if D.version_number>1:P=A.config_context.get_version_state_archive(D.version_number-1);H=create_tmp_archives_by_serialization_mechanism(P)
		else:H={}
		for (B,I) in G.items():
			try:
				J=CloudPodsMergeManager.get(B,raise_if_missing=_A);K=H.get(B,_B);C=O.get(B,_B)
				if not C:LOG.warning('No merge performed for %s serialized state files',B);continue
				elif K:J.three_way_merge(common_ancestor_state_files=K,from_state_files=C,to_state_files=I)
				else:J.two_way_merge(from_state_files=I,to_state_files=C)
			except Exception as Q:LOG.warning('Failed to perform merge for %s: %s',B,Q)
		shutil.make_archive(base_name=os.path.splitext(F)[0],format=COMPRESSION_FORMAT,root_dir=G['root'])
	def add_assets_to_pod(B,assets_root_paths):
		for A in assets_root_paths:
			C=os.path.join(B.config_context.get_assets_root_path(),os.path.basename(A))
			try:cp_r(src=A,dst=C)
			except Exception as D:LOG.warning('Failed to copy assets for %s: %s',A,D)
	def create_delta_log(A,state_from,state_to,diff_method=MetamodelDeltaMethod.SIMPLE):
		try:C=MetamodelDelta.get(diff_method);return C.create_delta_log(state_from,state_to,A.config_context)
		except Exception as D:LOG.debug('Unable to create delta log for version graph nodes: %s',D);B=short_uid();E=os.path.join(A.config_context.get_delta_log_path(),B);save_file(E,'{}');return B
	def upload_version_and_product_space(A,presigned_urls):
		D=presigned_urls;C='rb'
		def B(pre_signed_url,zip_data_content):
			A=safe_requests.put(pre_signed_url,data=zip_data_content)
			if A.status_code>=400:raise Exception(f"Unable to upload pod state to S3 (code {A.status_code}): {A.content}")
			return A
		H=D.get('presigned_version_space_url');E=A.create_version_space_archive()
		with open(E,C)as I:B(H,I.read())
		J=D.get('presigned_meta_state_urls');rm_rf(E)
		for (F,G) in J.items():
			K=G['meta'];L=A.config_context.get_version_meta_archive(F)
			with open(L,C)as M:B(K,M)
			N=G['state'];O=A.config_context.get_version_state_archive(F)
			with open(O,C)as P:B(N,P)